from __future__ import annotations

from pathlib import Path
from typing import Annotated, Final, TypeAlias

import msgspec
from construct import Array, Bytes, Int16ul, Int32ul, Struct

from ..game_modes import GameMode
from ..weapon_usage import (
    WEAPON_USAGE_SLOT_COUNT,
    ZERO_WEAPON_USAGE_COUNTS,
    WeaponUsageCounts,
    weapon_usage_slot_for_weapon_id,
)

GAME_CFG_NAME = "game.cfg"

BLOB_SIZE = 0x268
FILE_SIZE = BLOB_SIZE + 4

WEAPON_USAGE_COUNT = WEAPON_USAGE_SLOT_COUNT

# Quest play count length inferred from known trailing fields in the blob (0xD8..0x244).
QUEST_PLAY_COUNT = 91

UNKNOWN_TAIL_SIZE = 0x10

QuestPlayCounts: TypeAlias = Annotated[
    tuple[int, ...],
    msgspec.Meta(min_length=QUEST_PLAY_COUNT, max_length=QUEST_PLAY_COUNT),
]

_ZERO_QUEST_PLAY_COUNTS: Final[QuestPlayCounts] = tuple(0 for _ in range(QUEST_PLAY_COUNT))
_ZERO_UNKNOWN_TAIL: Final[bytes] = b"\x00" * UNKNOWN_TAIL_SIZE
_STATUS_FIELD_NAMES: Final[frozenset[str]] = frozenset(
    {
        "quest_unlock_index",
        "quest_unlock_index_full",
        "weapon_usage_counts",
        "quest_play_counts",
        "mode_play_survival",
        "mode_play_rush",
        "mode_play_typo",
        "mode_play_other",
        "game_sequence_id",
        "unknown_tail",
    },
)

GAME_STATUS_STRUCT = Struct(
    "quest_unlock_index" / Int16ul,
    "quest_unlock_index_full" / Int16ul,
    "weapon_usage_counts" / Array(WEAPON_USAGE_COUNT, Int32ul),
    "quest_play_counts" / Array(QUEST_PLAY_COUNT, Int32ul),
    "mode_play_survival" / Int32ul,
    "mode_play_rush" / Int32ul,
    "mode_play_typo" / Int32ul,
    "mode_play_other" / Int32ul,
    "game_sequence_id" / Int32ul,
    "unknown_tail" / Bytes(UNKNOWN_TAIL_SIZE),
)

GAME_CFG_STRUCT = Struct(
    "encoded" / Bytes(BLOB_SIZE),
    "checksum" / Int32ul,
)


class GameStatusData(msgspec.Struct):
    quest_unlock_index: int = 0
    quest_unlock_index_full: int = 0
    weapon_usage_counts: WeaponUsageCounts = msgspec.field(default_factory=lambda: ZERO_WEAPON_USAGE_COUNTS)
    quest_play_counts: QuestPlayCounts = msgspec.field(default_factory=lambda: _ZERO_QUEST_PLAY_COUNTS)
    mode_play_survival: int = 0
    mode_play_rush: int = 0
    mode_play_typo: int = 0
    mode_play_other: int = 0
    game_sequence_id: int = 0
    unknown_tail: bytes = _ZERO_UNKNOWN_TAIL


class GameStatus(GameStatusData, kw_only=True):
    path: Path
    dirty: bool = False

    def __setattr__(self, name: str, value: object) -> None:
        mark_dirty = name in _STATUS_FIELD_NAMES
        current = _MISSING
        if mark_dirty:
            try:
                current = object.__getattribute__(self, name)
            except AttributeError:
                pass
        super().__setattr__(name, value)
        if mark_dirty and current is not _MISSING and current != value:
            super().__setattr__("dirty", True)

    @classmethod
    def from_data(cls, *, path: Path, data: GameStatusData, dirty: bool = False) -> GameStatus:
        return cls(
            path=path,
            dirty=dirty,
            quest_unlock_index=data.quest_unlock_index,
            quest_unlock_index_full=data.quest_unlock_index_full,
            weapon_usage_counts=tuple(data.weapon_usage_counts),
            quest_play_counts=tuple(data.quest_play_counts),
            mode_play_survival=data.mode_play_survival,
            mode_play_rush=data.mode_play_rush,
            mode_play_typo=data.mode_play_typo,
            mode_play_other=data.mode_play_other,
            game_sequence_id=data.game_sequence_id,
            unknown_tail=bytes(data.unknown_tail),
        )

    def as_data(self) -> GameStatusData:
        return GameStatusData(
            quest_unlock_index=self.quest_unlock_index,
            quest_unlock_index_full=self.quest_unlock_index_full,
            weapon_usage_counts=tuple(self.weapon_usage_counts),
            quest_play_counts=tuple(self.quest_play_counts),
            mode_play_survival=self.mode_play_survival,
            mode_play_rush=self.mode_play_rush,
            mode_play_typo=self.mode_play_typo,
            mode_play_other=self.mode_play_other,
            game_sequence_id=self.game_sequence_id,
            unknown_tail=bytes(self.unknown_tail),
        )

    def mode_play_count_for_mode(self, game_mode: GameMode) -> int:
        return getattr(self, _mode_count_field_for_mode(game_mode))

    def increment_mode_play_count_for_mode(self, game_mode: GameMode, delta: int = 1) -> int:
        field = _mode_count_field_for_mode(game_mode)
        value = getattr(self, field) + int(delta)
        setattr(self, field, value)
        return value

    def weapon_usage_count_slot(self, slot: int) -> int:
        slot_idx = _require_index(slot, size=WEAPON_USAGE_COUNT, field="weapon_usage_slot")
        return int(self.weapon_usage_counts[slot_idx])

    def increment_weapon_usage_slot(self, slot: int, delta: int = 1) -> int:
        slot_idx = _require_index(slot, size=WEAPON_USAGE_COUNT, field="weapon_usage_slot")
        counts = list(self.weapon_usage_counts)
        counts[slot_idx] = counts[slot_idx] + int(delta)
        self.weapon_usage_counts = tuple(counts)
        return counts[slot_idx]

    def weapon_usage_count_for_weapon_id(self, weapon_id: int) -> int:
        slot = weapon_usage_slot_for_weapon_id(weapon_id)
        if slot is None:
            return 0
        return self.weapon_usage_count_slot(slot)

    def increment_weapon_usage_for_weapon_id(self, weapon_id: int, delta: int = 1) -> int | None:
        slot = weapon_usage_slot_for_weapon_id(weapon_id)
        if slot is None:
            return None
        return self.increment_weapon_usage_slot(slot, delta=delta)

    def quest_play_count(self, index: int) -> int:
        quest_idx = _require_index(index, size=QUEST_PLAY_COUNT, field="quest_play_count")
        return int(self.quest_play_counts[quest_idx])

    def increment_quest_play_count(self, index: int, delta: int = 1) -> int:
        quest_idx = _require_index(index, size=QUEST_PLAY_COUNT, field="quest_play_count")
        counts = list(self.quest_play_counts)
        counts[quest_idx] = counts[quest_idx] + int(delta)
        self.quest_play_counts = tuple(counts)
        return counts[quest_idx]

    def save(self) -> None:
        self.path.parent.mkdir(parents=True, exist_ok=True)
        save_status(self.path, self)
        self.dirty = False

    def save_if_dirty(self) -> None:
        if self.dirty:
            self.save()


class _Missing:
    pass


_MISSING: Final = _Missing()


def _mode_count_field_for_mode(game_mode: GameMode) -> str:
    mode = game_mode if isinstance(game_mode, GameMode) else GameMode(int(game_mode))
    match mode:
        case GameMode.SURVIVAL:
            return "mode_play_survival"
        case GameMode.RUSH:
            return "mode_play_rush"
        case GameMode.TYPO:
            return "mode_play_typo"
        case _:
            return "mode_play_other"


def _require_index(index: int, *, size: int, field: str) -> int:
    idx = int(index)
    if 0 <= idx < int(size):
        return idx
    raise IndexError(f"{field} out of range: {idx}")


def _status_blob_dict(data: GameStatusData) -> dict[str, object]:
    return {
        "quest_unlock_index": data.quest_unlock_index,
        "quest_unlock_index_full": data.quest_unlock_index_full,
        "weapon_usage_counts": list(data.weapon_usage_counts),
        "quest_play_counts": list(data.quest_play_counts),
        "mode_play_survival": data.mode_play_survival,
        "mode_play_rush": data.mode_play_rush,
        "mode_play_typo": data.mode_play_typo,
        "mode_play_other": data.mode_play_other,
        "game_sequence_id": data.game_sequence_id,
        "unknown_tail": bytes(data.unknown_tail),
    }


def default_status_data() -> GameStatusData:
    return GameStatusData()


def parse_status_blob(decoded: bytes) -> GameStatusData:
    if len(decoded) != BLOB_SIZE:
        raise ValueError(f"expected decoded blob of {BLOB_SIZE:#x} bytes, got {len(decoded):#x}")
    raw = GAME_STATUS_STRUCT.parse(decoded)
    return GameStatusData(
        quest_unlock_index=int(raw["quest_unlock_index"]),
        quest_unlock_index_full=int(raw["quest_unlock_index_full"]),
        weapon_usage_counts=tuple(int(value) for value in raw["weapon_usage_counts"]),
        quest_play_counts=tuple(int(value) for value in raw["quest_play_counts"]),
        mode_play_survival=int(raw["mode_play_survival"]),
        mode_play_rush=int(raw["mode_play_rush"]),
        mode_play_typo=int(raw["mode_play_typo"]),
        mode_play_other=int(raw["mode_play_other"]),
        game_sequence_id=int(raw["game_sequence_id"]),
        unknown_tail=bytes(raw["unknown_tail"]),
    )


def build_status_blob(data: GameStatusData) -> bytes:
    return GAME_STATUS_STRUCT.build(_status_blob_dict(data))


def to_s8(value: int) -> int:
    value &= 0xFF
    return value - 0x100 if value & 0x80 else value


def index_poly(idx: int) -> int:
    i = to_s8(idx)
    return ((i * 7 + 0x0F) * i + 0x03) * i


def decode_blob(encoded: bytes) -> bytes:
    if len(encoded) != BLOB_SIZE:
        raise ValueError(f"decoded blob must be {BLOB_SIZE:#x} bytes, got {len(encoded):#x}")
    decoded = bytearray(encoded)
    for i in range(BLOB_SIZE):
        decoded[i] = (decoded[i] - 0x6F - index_poly(i)) & 0xFF
    return bytes(decoded)


def encode_blob(decoded: bytes) -> bytes:
    if len(decoded) != BLOB_SIZE:
        raise ValueError(f"decoded blob must be {BLOB_SIZE:#x} bytes, got {len(decoded):#x}")
    encoded = bytearray(decoded)
    for i in range(BLOB_SIZE):
        encoded[i] = (encoded[i] + 0x6F + index_poly(i)) & 0xFF
    return bytes(encoded)


def compute_checksum(decoded: bytes) -> int:
    acc = 0
    u = 0
    for i, b in enumerate(decoded):
        c = to_s8(b)
        i_var5 = (c * 7 + i) * c + u
        acc = (acc + 0x0D + i_var5) & 0xFFFFFFFF
        u += 0x6F
    return acc


def load_status(path: Path) -> GameStatus:
    raw = path.read_bytes()
    if len(raw) != FILE_SIZE:
        raise ValueError(f"expected {FILE_SIZE:#x} bytes, got {len(raw):#x}")
    parsed = GAME_CFG_STRUCT.parse(raw)
    encoded = bytes(parsed["encoded"])
    stored_checksum = int(parsed["checksum"])
    decoded = decode_blob(encoded)
    computed = compute_checksum(decoded)
    if stored_checksum != computed:
        raise ValueError("checksum mismatch")
    return GameStatus.from_data(path=path, data=parse_status_blob(decoded), dirty=False)


def save_status(path: Path, status: GameStatusData | GameStatus) -> None:
    decoded = build_status_blob(status.as_data() if isinstance(status, GameStatus) else status)
    checksum = compute_checksum(decoded)
    encoded = encode_blob(decoded)
    path.write_bytes(GAME_CFG_STRUCT.build({"encoded": encoded, "checksum": checksum}))


def ensure_game_status(base_dir: Path) -> GameStatus:
    path = base_dir / GAME_CFG_NAME
    if path.exists():
        return load_status(path)
    status = GameStatus.from_data(path=path, data=default_status_data(), dirty=False)
    status.save()
    return status


def hash_status_data(status: GameStatusData) -> str:
    import hashlib

    return hashlib.sha256(build_status_blob(status)).hexdigest()


__all__ = [
    "BLOB_SIZE",
    "FILE_SIZE",
    "GAME_CFG_NAME",
    "GAME_CFG_STRUCT",
    "GAME_STATUS_STRUCT",
    "GameStatus",
    "GameStatusData",
    "QUEST_PLAY_COUNT",
    "QuestPlayCounts",
    "UNKNOWN_TAIL_SIZE",
    "WEAPON_USAGE_COUNT",
    "build_status_blob",
    "compute_checksum",
    "decode_blob",
    "default_status_data",
    "encode_blob",
    "ensure_game_status",
    "hash_status_data",
    "load_status",
    "parse_status_blob",
    "save_status",
]



from __future__ import annotations

import hashlib
import platform
import subprocess
from datetime import UTC, datetime
from pathlib import Path
from typing import Literal

import msgspec

from grim.rand import CallerStatic

from ..replay import load_replay_file
from ..replay.checkpoints import ReplayCheckpoint
from ..replay.driver.playback_driver import PlaybackWalkHooks, build_verify_playback_driver
from ..replay.types import Replay
from ..sim.timing import ftol_ms_i32
from ..sim.world_state import WorldState
from .canonical_channels import (
    BonusEntitySample,
    CreatureEntitySample,
    EntitySamplesSnapshot,
    ProjectileEntitySample,
    RngStreamRow,
    SecondaryProjectileEntitySample,
    SimStateSnapshot,
    SnapshotBonusTimers,
    SnapshotGameplay,
    SnapshotPlayer,
    SnapshotVec2,
    SnapshotWeapon,
    bonus_timer_ms,
)
from .payloads import BuiltinObject
from .schema import (
    TRACE_FORMAT_VERSION,
    TRACE_REQUIRED_CHANNELS,
    TRACE_SCHEMA_VERSION,
    ReplayTickChannels,
    TickRecord,
    TraceMeta,
    channel_versions_for,
)
from .trace import TraceError, TraceReader, TraceSummary, write_trace

_REPO_ROOT = Path(__file__).resolve().parents[3]
_ZIG_ROOT = _REPO_ROOT / "crimson-zig"
_ZIG_BIN = _ZIG_ROOT / "zig-out" / "bin" / "crimson-zig"


class _EntityGenerationState(msgspec.Struct):
    generation_by_index: dict[int, int] = msgspec.field(default_factory=dict)
    active_indices: set[int] = msgspec.field(default_factory=set)
    _seen_in_tick: set[int] = msgspec.field(default_factory=set)

    def begin_tick(self) -> None:
        self._seen_in_tick.clear()

    def end_tick(self) -> None:
        self.active_indices = set(self._seen_in_tick)

    def next_generation(self, *, index: int) -> int:
        idx = int(index)
        if idx not in self.generation_by_index:
            self.generation_by_index[idx] = 0
        if idx not in self.active_indices:
            self.generation_by_index[idx] += 1
        self._seen_in_tick.add(idx)
        return int(self.generation_by_index[idx])


def _fingerprint(path: Path) -> BuiltinObject:
    stat = path.stat()
    raw = path.read_bytes()
    return {
        "path": str(path),
        "sha256": hashlib.sha256(raw).hexdigest(),
        "size": stat.st_size,
        "mtime_ns": stat.st_mtime_ns,
    }


def _rng_stream_from_draws(draws: list[tuple[int, int, int, CallerStatic]]) -> list[RngStreamRow]:
    rows: list[RngStreamRow] = []
    for index, row in enumerate(draws):
        state_before_u32, value_15, state_after_u32, caller = row
        rows.append(
            RngStreamRow(
                tick_call_index=int(index) + 1,
                value_15=int(value_15),
                state_before_u32=int(state_before_u32),
                state_after_u32=int(state_after_u32),
                caller=caller,
            ),
        )
    return rows


def _entity_samples_for_world(
    world: WorldState,
    *,
    creature_state: _EntityGenerationState,
    projectile_state: _EntityGenerationState,
    secondary_state: _EntityGenerationState,
    bonus_state: _EntityGenerationState,
) -> EntitySamplesSnapshot:
    creature_state.begin_tick()
    projectile_state.begin_tick()
    secondary_state.begin_tick()
    bonus_state.begin_tick()

    creatures: list[CreatureEntitySample] = []
    for index, creature in enumerate(world.creatures.entries):
        if not creature.active:
            continue
        generation = creature_state.next_generation(index=index)
        creatures.append(
            CreatureEntitySample(
                uid=int(index),
                generation=generation,
                pool_kind="creature",
                index=index,
                active=True,
                type_id=int(creature.type_id),
                hp=float(creature.hp),
                pos=SnapshotVec2(x=float(creature.pos.x), y=float(creature.pos.y)),
                flags=int(creature.flags),
                ai_mode=int(creature.ai_mode),
                link_index=int(creature.link_index),
                heading=float(creature.heading),
                target_heading=float(creature.target_heading),
                orbit_angle=float(creature.orbit_angle),
                orbit_radius=float(creature.orbit_radius),
                lifecycle_stage=float(creature.lifecycle_stage),
            ),
        )

    projectiles: list[ProjectileEntitySample] = []
    for index, projectile in enumerate(world.state.projectiles.entries):
        if not projectile.active:
            continue
        generation = projectile_state.next_generation(index=index)
        projectiles.append(
            ProjectileEntitySample(
                uid=int(index),
                generation=generation,
                pool_kind="projectile",
                index=index,
                active=True,
                type_id=int(projectile.type_id),
                angle=float(projectile.angle),
                pos=SnapshotVec2(x=float(projectile.pos.x), y=float(projectile.pos.y)),
                vel=SnapshotVec2(x=float(projectile.vel.x), y=float(projectile.vel.y)),
                life_timer=float(projectile.life_timer),
                speed_scale=float(projectile.speed_scale),
                damage_pool=float(projectile.damage_pool),
                hit_radius=float(projectile.hit_radius),
                travel_budget=float(projectile.travel_budget),
                owner_id=int(projectile.owner.to_legacy()),
            ),
        )

    secondary_projectiles: list[SecondaryProjectileEntitySample] = []
    for index, projectile in enumerate(world.state.secondary_projectiles.entries):
        if not projectile.active:
            continue
        generation = secondary_state.next_generation(index=index)
        secondary_projectiles.append(
            SecondaryProjectileEntitySample(
                uid=int(index),
                generation=generation,
                pool_kind="secondary_projectile",
                index=index,
                active=True,
                type_id=int(projectile.type_id),
                angle=float(projectile.angle),
                pos=SnapshotVec2(x=float(projectile.pos.x), y=float(projectile.pos.y)),
                vel=SnapshotVec2(x=float(projectile.vel.x), y=float(projectile.vel.y)),
                speed=float(projectile.speed),
                trail_timer=float(projectile.trail_timer),
                owner_id=int(projectile.owner.to_legacy()),
                target_id=int(projectile.target_id),
            ),
        )

    bonuses: list[BonusEntitySample] = []
    for index, bonus in enumerate(world.state.bonus_pool.entries):
        if int(bonus.bonus_id) == 0:
            continue
        generation = bonus_state.next_generation(index=index)
        bonuses.append(
            BonusEntitySample(
                uid=int(index),
                generation=generation,
                pool_kind="bonus",
                index=index,
                active=True,
                bonus_id=int(bonus.bonus_id),
                picked=bool(bonus.picked),
                time_left=float(bonus.time_left),
                time_max=float(bonus.time_max),
                pos=SnapshotVec2(x=float(bonus.pos.x), y=float(bonus.pos.y)),
                amount=int(bonus.amount),
            ),
        )

    creature_state.end_tick()
    projectile_state.end_tick()
    secondary_state.end_tick()
    bonus_state.end_tick()

    return EntitySamplesSnapshot(
        creatures=creatures,
        projectiles=projectiles,
        secondary_projectiles=secondary_projectiles,
        bonuses=bonuses,
    )
def _sim_state_from_world(world: WorldState, *, replay: Replay) -> SimStateSnapshot:
    gameplay = world.state
    players: list[SnapshotPlayer] = []
    for player in world.players:
        players.append(
            SnapshotPlayer(
                index=int(player.index),
                pos=SnapshotVec2(x=float(player.pos.x), y=float(player.pos.y)),
                health=float(player.health),
                weapon=SnapshotWeapon(
                    weapon_id=int(player.weapon.weapon_id),
                    ammo=float(player.weapon.ammo),
                    clip_size=int(player.weapon.clip_size),
                    reload_active=bool(player.weapon.reload_active),
                    reload_timer=float(player.weapon.reload_timer),
                    reload_timer_max=float(player.weapon.reload_timer_max),
                    shot_cooldown=float(player.weapon.shot_cooldown),
                ),
                experience=int(player.experience),
                level=int(player.level),
            ),
        )
    return SimStateSnapshot(
        gameplay=SnapshotGameplay(
            mode_id=int(replay.header.game_mode_id),
            quest_stage_major=(0 if gameplay.quest_level is None else int(gameplay.quest_level.major)),
            quest_stage_minor=(0 if gameplay.quest_level is None else int(gameplay.quest_level.minor)),
            perk_pending_count=int(gameplay.perk_selection.pending_count),
            perk_choices_dirty=bool(gameplay.perk_selection.choices_dirty),
            bonus_timers=SnapshotBonusTimers(
                weapon_power_up_ms=bonus_timer_ms(float(gameplay.bonuses.weapon_power_up)),
                reflex_boost_ms=bonus_timer_ms(float(gameplay.bonuses.reflex_boost)),
                energizer_ms=bonus_timer_ms(float(gameplay.bonuses.energizer)),
                double_experience_ms=bonus_timer_ms(float(gameplay.bonuses.double_experience)),
                freeze_ms=bonus_timer_ms(float(gameplay.bonuses.freeze)),
            ),
        ),
        players=players,
    )


def _build_replay_fingerprint(*, replay_path: Path, replay: Replay) -> BuiltinObject:
    replay_fingerprint = _fingerprint(replay_path)
    replay_fingerprint["tick_rate"] = replay.header.tick_rate
    replay_fingerprint["seed"] = replay.header.seed
    replay_fingerprint["mode_id"] = replay.header.game_mode_id
    replay_fingerprint["quest_level"] = (
        ""
        if replay.header.quest_level is None
        else replay.header.quest_level.text
    )
    return replay_fingerprint


def _build_trace_meta(
    *,
    replay_path: Path,
    replay: Replay,
    tick_rows: list[TickRecord],
    channels_seen: set[str],
    impl: Literal["python", "zig"],
    config_extra: BuiltinObject | None = None,
) -> TraceMeta:
    tick_start = min((row.tick_index for row in tick_rows), default=-1)
    tick_end = max((row.tick_index for row in tick_rows), default=-1)
    replay_fingerprint = _build_replay_fingerprint(replay_path=replay_path, replay=replay)
    channels_sorted = sorted(channels_seen)
    config = {
        "impl": str(impl),
    }
    if config_extra is not None:
        config.update(config_extra)
    return TraceMeta(
        trace_format_version=TRACE_FORMAT_VERSION,
        trace_schema_version=TRACE_SCHEMA_VERSION,
        created_utc=datetime.now(tz=UTC).isoformat(),
        producer={
            "impl": str(impl),
            "impl_version": "",
            "platform": str(platform.system()),
            "arch": str(platform.machine()),
        },
        source=replay_fingerprint,
        channels=channels_sorted,
        channel_versions=channel_versions_for(channels_sorted),
        tick_range={
            "start_tick": tick_start,
            "end_tick": tick_end,
            "tick_count": len(tick_rows),
        },
        config=config,
        status=replay.header.status,
    )


def _record_replay_to_trace_python(
    *,
    replay_path: Path,
    out_path: Path,
    chunk_ticks: int,
) -> TraceSummary:
    replay = load_replay_file(replay_path)

    replay_tick_count = len(replay.ticks)
    checkpoint_ticks = set(range(replay_tick_count))
    checkpoints: list[ReplayCheckpoint] = []

    entity_samples_by_tick: dict[int, EntitySamplesSnapshot] = {}
    sim_state_by_tick: dict[int, SimStateSnapshot] = {}
    rng_stream_by_tick: dict[int, list[RngStreamRow]] = {}
    creature_state = _EntityGenerationState()
    projectile_state = _EntityGenerationState()
    secondary_state = _EntityGenerationState()
    bonus_state = _EntityGenerationState()

    def _tick_observer(tick_index: int, world: WorldState) -> None:
        entity_samples_by_tick[tick_index] = _entity_samples_for_world(
            world,
            creature_state=creature_state,
            projectile_state=projectile_state,
            secondary_state=secondary_state,
            bonus_state=bonus_state,
        )
        sim_state_by_tick[tick_index] = _sim_state_from_world(world, replay=replay)

    def _tick_rng_trace_observer(tick_index: int, draws: list[tuple[int, int, int, CallerStatic]]) -> None:
        rng_stream_by_tick[int(tick_index)] = _rng_stream_from_draws(draws)

    driver = build_verify_playback_driver(
        replay,
        max_ticks=None,
        trace_rng=True,
        strict_rng_trace=False,
    )

    def _after_tick(tick_result, world: WorldState) -> None:
        tick_index = int(tick_result.source_tick.tick_index)
        if tick_index in checkpoint_ticks:
            checkpoints.append(driver.build_checkpoint(tick_result=tick_result))
        _tick_observer(tick_index, world)

    def _on_rng_trace(tick_result, draws: tuple[tuple[int, int, int, CallerStatic], ...]) -> None:
        _tick_rng_trace_observer(
            int(tick_result.source_tick.tick_index),
            list(draws),
        )

    driver.run(
        hooks=PlaybackWalkHooks(
            after_tick=_after_tick,
            on_rng_trace=_on_rng_trace,
        ),
    )

    tick_rows: list[TickRecord] = []
    channels_seen: set[str] = set()
    replay_dt_rows = [ftol_ms_i32(tick.dt) for tick in replay.ticks]
    for checkpoint in sorted(checkpoints, key=lambda row: row.tick_index):
        tick_index = int(checkpoint.tick_index)
        if tick_index not in rng_stream_by_tick:
            raise ValueError(f"missing runtime rng stream for tick {tick_index}")
        if tick_index not in entity_samples_by_tick:
            raise ValueError(f"missing entity_samples snapshot for tick {tick_index}")
        if tick_index not in sim_state_by_tick:
            raise ValueError(f"missing sim_state snapshot for tick {tick_index}")

        entity_samples_obj = entity_samples_by_tick[tick_index]
        sim_state_obj = sim_state_by_tick[tick_index]
        rng_stream = list(rng_stream_by_tick[tick_index])

        channels = ReplayTickChannels(
            checkpoint=checkpoint,
            sim_state=sim_state_obj,
            entity_samples=entity_samples_obj,
            rng_stream=rng_stream,
            timing_samples=[],
        )

        if not (0 <= tick_index < len(replay_dt_rows)):
            raise ValueError(f"missing replay dt_ms_i32 row for tick {tick_index}")
        tick_dt_ms_i32 = int(replay_dt_rows[tick_index])
        if tick_dt_ms_i32 < 0:
            raise ValueError(f"invalid replay dt_ms_i32 at tick {tick_index}: {tick_dt_ms_i32}")

        channels_seen.update(TRACE_REQUIRED_CHANNELS)
        tick_rows.append(
            TickRecord(
                tick_index=tick_index,
                elapsed_ms=int(checkpoint.elapsed_ms),
                dt_ms_i32=tick_dt_ms_i32,
                mode_id=int(replay.header.game_mode_id),
                phase_markers=[],
                channels=channels,
            ),
        )

    meta = _build_trace_meta(
        replay_path=replay_path,
        replay=replay,
        tick_rows=tick_rows,
        channels_seen=channels_seen,
        impl="python",
        config_extra=None,
    )
    return write_trace(
        out_path,
        meta=meta,
        ticks=tick_rows,
        chunk_ticks=max(1, chunk_ticks),
    )


def _command_detail(run: subprocess.CompletedProcess[str]) -> str:
    stderr = str(run.stderr).strip()
    stdout = str(run.stdout).strip()
    if stderr:
        return stderr
    if stdout:
        return stdout
    return "(no command output)"


def _run_process(command: list[str], *, cwd: Path) -> subprocess.CompletedProcess[str]:
    try:
        return subprocess.run(
            command,
            cwd=cwd,
            check=False,
            text=True,
            capture_output=True,
        )
    except OSError as exc:
        joined = " ".join(command)
        raise ValueError(f"failed to run command: {joined}: {exc}") from exc


def _record_replay_to_trace_zig(
    *,
    replay_path: Path,
    out_path: Path,
    chunk_ticks: int,
) -> tuple[TraceSummary, list[str]]:
    replay = load_replay_file(replay_path)
    if int(replay.header.player_count) != 1:
        raise ValueError(
            "zig dbg record currently supports only single-player replays; "
            f"got player_count={int(replay.header.player_count)}",
        )

    build_run = _run_process(["zig", "build"], cwd=_ZIG_ROOT)
    if int(build_run.returncode) != 0:
        raise ValueError(
            "zig build failed: "
            f"exit={int(build_run.returncode)} detail={_command_detail(build_run)}",
        )

    warnings: list[str] = []
    verify_cmd = [
        str(_ZIG_BIN),
        "replay",
        "verify",
        str(replay_path),
        "--debug-trace-cdt",
        str(out_path),
        "--debug-trace-cdt-chunk-ticks",
        str(max(1, int(chunk_ticks))),
        "--format",
        "json",
    ]
    verify_run = _run_process(verify_cmd, cwd=_REPO_ROOT)
    if not out_path.is_file():
        raise ValueError(
            "zig trace generation failed: replay verify did not produce CDT output; "
            f"exit={int(verify_run.returncode)} detail={_command_detail(verify_run)}",
        )
    if int(verify_run.returncode) != 0:
        warning = (
            f"warning: zig replay verify exited {int(verify_run.returncode)}; "
            "continuing with emitted trace"
        )
        stderr = str(verify_run.stderr).strip()
        if stderr:
            warning = f"{warning}: {stderr.splitlines()[0]}"
        warnings.append(warning)

    try:
        with TraceReader(out_path) as trace:
            decoded_tick_count = sum(1 for _ in trace.iter_ticks())
            if int(decoded_tick_count) != int(trace.footer.tick_count):
                raise TraceError(
                    "zig trace validation failed to decode all tick payloads: "
                    f"decoded={int(decoded_tick_count)} footer={int(trace.footer.tick_count)}",
                )
            summary = TraceSummary(meta=trace.meta, footer=trace.footer)
    except TraceError as exc:
        raise ValueError(f"zig trace validation failed: {exc}") from exc
    return summary, warnings


def record_replay_to_trace(
    *,
    replay_path: Path,
    out_path: Path,
    impl: Literal["python", "zig"] = "python",
    chunk_ticks: int = 256,
    warnings_out: list[str] | None = None,
) -> TraceSummary:
    replay_path = Path(replay_path)
    out_path = Path(out_path)
    if warnings_out is None:
        warnings_out = []
    if str(impl) == "python":
        summary = _record_replay_to_trace_python(
            replay_path=replay_path,
            out_path=out_path,
            chunk_ticks=chunk_ticks,
        )
        return summary
    if str(impl) == "zig":
        summary, warnings = _record_replay_to_trace_zig(
            replay_path=replay_path,
            out_path=out_path,
            chunk_ticks=chunk_ticks,
        )
        warnings_out.extend(warnings)
        return summary
    raise ValueError(f"unsupported dbg record impl: {impl!r}")



from __future__ import annotations

import io
from collections.abc import Sequence
from pathlib import Path

import msgspec
import zstandard as zstd

from grim.geom import Vec2

from ..bonuses import BonusId
from ..creatures.runtime import CreatureDeath
from ..game_modes import GameMode
from ..sim.state_types import PlayerState
from ..sim.timing import ftol_ms_i32
from ..sim.world_state import WorldEvents, WorldState
from ..weapons import WeaponId

FORMAT_VERSION = 4
DEFAULT_CHECKPOINT_SAMPLE_RATE = 1
_ZSTD_MAGIC = b"\x28\xb5\x2f\xfd"
_DEFAULT_MAX_CHECKPOINTS_PAYLOAD_BYTES = 256 * 1024 * 1024
_CHECKPOINTS_ZSTD_LEVEL = 19


class ReplayCheckpointsError(ValueError):
    pass


class ReplayPlayerCheckpoint(msgspec.Struct, frozen=True):
    pos: Vec2
    health: float
    weapon_id: WeaponId
    ammo: float
    experience: int
    level: int


class ReplayTypoNameEntry(msgspec.Struct, frozen=True):
    creature_index: int
    name: str


class ReplayTypoSnapshot(msgspec.Struct, frozen=True):
    input_text: str = ""
    submit_count: int = 0
    match_count: int = 0
    spawn_cooldown_ms: int = 0
    active_names: list[ReplayTypoNameEntry] = msgspec.field(default_factory=list)


class ReplayTutorialSnapshot(msgspec.Struct, frozen=True):
    stage_index: int = -1
    stage_timer_ms: int = 0
    stage_transition_timer_ms: int = -1000
    hint_index: int = -1
    hint_alpha: int = 0
    hint_fade_in: bool = False
    repeat_spawn_count: int = 0
    hint_bonus_creature_ref: int | None = None
    prompt_text: str = ""
    prompt_alpha: float = 0.0
    hint_text: str = ""
    hint_alpha_overlay: float = 0.0


class ReplayCheckpoint(msgspec.Struct, frozen=True):
    tick_index: int
    rng_state: int
    elapsed_ms: int
    score_xp: int
    kills: int
    creature_count: int
    perk_pending: int
    players: list[ReplayPlayerCheckpoint]
    bonus_timers: dict[str, int]
    deaths: list["ReplayDeathLedgerEntry"] = msgspec.field(default_factory=list)
    perk: "ReplayPerkSnapshot" = msgspec.field(default_factory=lambda: ReplayPerkSnapshot())
    events: "ReplayEventSummary" = msgspec.field(default_factory=lambda: ReplayEventSummary())
    tutorial: ReplayTutorialSnapshot | None = None
    typo: ReplayTypoSnapshot | None = None


class ReplayDeathLedgerEntry(msgspec.Struct, frozen=True):
    creature_index: int
    type_id: int
    reward_value: float
    xp_awarded: int
    owner_id: int


class ReplayPerkSnapshot(msgspec.Struct, frozen=True):
    pending_count: int = 0
    choices_dirty: bool = False
    choices: list[int] = msgspec.field(default_factory=list)
    player_nonzero_counts: list[list[list[int]]] = msgspec.field(default_factory=list)


class ReplayEventSummary(msgspec.Struct, frozen=True):
    # Legacy sidecars may omit this block; -1 marks "unknown/not recorded".
    hit_count: int = 0
    pickup_count: int = 0
    sfx_count: int = 0
    sfx_head: list[str] = msgspec.field(default_factory=list)


class ReplayCheckpoints(msgspec.Struct, frozen=True):
    version: int
    sample_rate: int
    checkpoints: list[ReplayCheckpoint] = msgspec.field(default_factory=list)

_CHECKPOINTS_ENCODER = msgspec.msgpack.Encoder()
_CHECKPOINTS_DECODER = msgspec.msgpack.Decoder(type=ReplayCheckpoints)


def default_checkpoints_path(replay_path: Path) -> Path:
    replay_path = Path(replay_path)
    return replay_path.with_name(f"{replay_path.name}.chk")


def _is_zstd(data: bytes) -> bool:
    return bytes(data).startswith(_ZSTD_MAGIC)


def _decompress_zstd_checkpoints(data: bytes, *, max_output_bytes: int) -> bytes:
    try:
        with zstd.ZstdDecompressor().stream_reader(io.BytesIO(data)) as stream:
            payload = stream.read(int(max_output_bytes) + 1)
    except zstd.ZstdError as exc:
        raise ReplayCheckpointsError("invalid checkpoints zstd payload") from exc
    if len(payload) > int(max_output_bytes):
        raise ReplayCheckpointsError(
            f"checkpoints payload too large after zstd decompression (> {int(max_output_bytes)} bytes)",
        )
    return payload


def _bonus_timer_ms(value: float) -> int:
    # Keep checkpoint values compact/stable: ms resolution is enough for divergence detection.
    ms = int(ftol_ms_i32(float(value)))
    if ms < 0:
        return 0
    return ms


def build_checkpoint(
    *,
    tick_index: int,
    world: WorldState,
    elapsed_ms: float,
    creature_count_override: int | None = None,
    deaths: Sequence[CreatureDeath] | None = None,
    events: WorldEvents | None = None,
) -> ReplayCheckpoint:
    state = world.state
    players: list[PlayerState] = list(world.players)
    score_xp = sum(int(player.experience) for player in players)
    kills = int(world.creatures.kill_count)
    if creature_count_override is None:
        creature_count = sum(1 for creature in world.creatures.entries if creature.active)
    else:
        creature_count = int(creature_count_override)

    player_ckpts: list[ReplayPlayerCheckpoint] = []
    for player in players:
        player_ckpts.append(
            ReplayPlayerCheckpoint(
                pos=Vec2(round(player.pos.x, 4), round(player.pos.y, 4)),
                health=round(player.health, 4),
                weapon_id=WeaponId(player.weapon.weapon_id),
                ammo=round(player.weapon.ammo, 4),
                experience=int(player.experience),
                level=int(player.level),
            ),
        )

    bonus_timers = {
        str(BonusId.WEAPON_POWER_UP): _bonus_timer_ms(state.bonuses.weapon_power_up),
        str(BonusId.REFLEX_BOOST): _bonus_timer_ms(state.bonuses.reflex_boost),
        str(BonusId.ENERGIZER): _bonus_timer_ms(state.bonuses.energizer),
        str(BonusId.DOUBLE_EXPERIENCE): _bonus_timer_ms(state.bonuses.double_experience),
        str(BonusId.FREEZE): _bonus_timer_ms(state.bonuses.freeze),
    }

    perk_counts: list[list[list[int]]] = []
    for player in players:
        nonzero: list[list[int]] = []
        for perk_id, count in enumerate(player.perk_counts):
            count_i = int(count)
            if count_i != 0:
                nonzero.append([int(perk_id), count_i])
        perk_counts.append(nonzero)

    perk_snapshot = ReplayPerkSnapshot(
        pending_count=int(state.perk_selection.pending_count),
        choices_dirty=bool(state.perk_selection.choices_dirty),
        choices=[int(perk_id) for perk_id in state.perk_selection.choices],
        player_nonzero_counts=perk_counts,
    )

    death_entries: list[ReplayDeathLedgerEntry] = []
    for death in deaths or ():
        death_entries.append(
            ReplayDeathLedgerEntry(
                creature_index=int(death.index),
                type_id=int(death.type_id),
                reward_value=float(death.reward_value),
                xp_awarded=int(death.xp_awarded),
                owner_id=int(death.owner.to_legacy()),
            ),
        )

    hits = list(events.hits) if events is not None else []
    pickups = list(events.pickups) if events is not None else []
    sfx = list(events.sfx) if events is not None else []
    event_summary = ReplayEventSummary(
        hit_count=len(hits),
        pickup_count=len(pickups),
        sfx_count=len(sfx),
        sfx_head=[key.value for key in sfx[:4]],
    )

    typo_snapshot: ReplayTypoSnapshot | None = None
    tutorial_snapshot: ReplayTutorialSnapshot | None = None
    if state.game_mode == GameMode.TUTORIAL:
        tutorial = state.tutorial
        overlay = state.tutorial_overlay
        tutorial_snapshot = ReplayTutorialSnapshot(
            stage_index=int(tutorial.stage_index),
            stage_timer_ms=int(tutorial.stage_timer_ms),
            stage_transition_timer_ms=int(tutorial.stage_transition_timer_ms),
            hint_index=int(tutorial.hint_index),
            hint_alpha=int(tutorial.hint_alpha),
            hint_fade_in=bool(tutorial.hint_fade_in),
            repeat_spawn_count=int(tutorial.repeat_spawn_count),
            hint_bonus_creature_ref=(
                None if tutorial.hint_bonus_creature_ref is None else int(tutorial.hint_bonus_creature_ref)
            ),
            prompt_text=str(overlay.prompt_text),
            prompt_alpha=float(overlay.prompt_alpha),
            hint_text=str(overlay.hint_text),
            hint_alpha_overlay=float(overlay.hint_alpha),
        )
    if state.game_mode == GameMode.TYPO:
        active_mask = [bool(creature.active) for creature in world.creatures.entries]
        active_names = [
            ReplayTypoNameEntry(
                creature_index=int(creature_index),
                name=str(name),
            )
            for creature_index, name in state.typo.names.active_entries(active_mask=active_mask)
        ]
        typo_snapshot = ReplayTypoSnapshot(
            input_text=str(state.typo.typing.text),
            submit_count=int(state.typo.typing.submit_count),
            match_count=int(state.typo.typing.match_count),
            spawn_cooldown_ms=int(state.typo.spawn_cooldown_ms),
            active_names=active_names,
        )

    return ReplayCheckpoint(
        tick_index=int(tick_index),
        rng_state=int(state.rng.state),
        elapsed_ms=int(round(elapsed_ms)),
        score_xp=int(score_xp),
        kills=int(kills),
        creature_count=int(creature_count),
        perk_pending=int(state.perk_selection.pending_count),
        players=player_ckpts,
        bonus_timers=bonus_timers,
        deaths=death_entries,
        perk=perk_snapshot,
        events=event_summary,
        tutorial=tutorial_snapshot,
        typo=typo_snapshot,
    )


def dump_checkpoints(checkpoints: ReplayCheckpoints) -> bytes:
    raw = _CHECKPOINTS_ENCODER.encode(checkpoints)
    return zstd.ZstdCompressor(level=_CHECKPOINTS_ZSTD_LEVEL).compress(raw)


def load_checkpoints(data: bytes) -> ReplayCheckpoints:
    max_payload_bytes = int(_DEFAULT_MAX_CHECKPOINTS_PAYLOAD_BYTES)
    payload = bytes(data)
    if _is_zstd(payload):
        payload = _decompress_zstd_checkpoints(payload, max_output_bytes=max_payload_bytes)
    if len(payload) > int(max_payload_bytes):
        raise ReplayCheckpointsError(f"checkpoints payload too large (> {int(max_payload_bytes)} bytes)")
    try:
        decoded = _CHECKPOINTS_DECODER.decode(payload)
    except (msgspec.DecodeError, msgspec.ValidationError) as exc:
        raise ReplayCheckpointsError("invalid checkpoints msgpack payload") from exc

    if int(decoded.version) != FORMAT_VERSION:
        raise ReplayCheckpointsError(f"unsupported checkpoints version: {int(decoded.version)}")
    return decoded


def dump_checkpoints_file(path: Path, checkpoints: ReplayCheckpoints) -> None:
    Path(path).write_bytes(dump_checkpoints(checkpoints))


def load_checkpoints_file(path: Path) -> ReplayCheckpoints:
    return load_checkpoints(Path(path).read_bytes())



from __future__ import annotations

from .databases_perks import UnlockedPerksDatabaseView
from .databases_weapons import UnlockedWeaponsDatabaseView

__all__ = [
    "UnlockedPerksDatabaseView",
    "UnlockedWeaponsDatabaseView",
]



from __future__ import annotations

import ast
import json
from collections import Counter, defaultdict
from dataclasses import dataclass
from pathlib import Path


@dataclass(frozen=True)
class StructClass:
    class_name: str
    full_name: str
    module: str
    path: str
    lineno: int


@dataclass(frozen=True)
class InventorySummary:
    total_structs: int
    counts_by_bucket: dict[str, int]
    duplicate_names: dict[str, list[StructClass]]


def _module_name_for_path(*, source_root: Path, py_path: Path) -> str:
    rel = py_path.relative_to(source_root)
    parts = list(rel.parts)
    if parts and parts[-1].endswith('.py'):
        parts[-1] = parts[-1][:-3]
    if parts and parts[-1] == '__init__':
        parts = parts[:-1]
    return '.'.join(part for part in parts if part)


def _resolve_import_module(*, current_module: str, imported_module: str | None, level: int) -> str:
    if level <= 0:
        return str(imported_module or '')

    parts = [part for part in current_module.split('.') if part]
    base_parts = parts[:-level] if level <= len(parts) else []
    if imported_module:
        base_parts.extend([part for part in imported_module.split('.') if part])
    return '.'.join(base_parts)


def _base_name(base: ast.expr) -> str | None:
    if isinstance(base, ast.Name):
        return base.id
    if isinstance(base, ast.Attribute):
        chain: list[str] = []
        cur: ast.expr | None = base
        while isinstance(cur, ast.Attribute):
            chain.append(cur.attr)
            cur = cur.value
        if isinstance(cur, ast.Name):
            chain.append(cur.id)
            return '.'.join(reversed(chain))
    return None


def _resolve_symbol(symbol: str, *, imports: dict[str, str]) -> str:
    if symbol in imports:
        return imports[symbol]
    if '.' not in symbol:
        return symbol
    root, rest = symbol.split('.', 1)
    if root in imports:
        return f"{imports[root]}.{rest}"
    return symbol


@dataclass(frozen=True)
class _ClassInfo:
    class_name: str
    full_name: str
    module: str
    path: str
    lineno: int
    resolved_bases: tuple[str, ...]


def _iter_class_infos(source_root: Path) -> list[_ClassInfo]:
    infos: list[_ClassInfo] = []
    for py_path in sorted(source_root.rglob('*.py')):
        module = _module_name_for_path(source_root=source_root, py_path=py_path)
        try:
            tree = ast.parse(py_path.read_text(encoding='utf-8'))
        except (SyntaxError, UnicodeDecodeError, OSError):
            continue

        imports: dict[str, str] = {}
        for node in tree.body:
            if isinstance(node, ast.Import):
                for alias in node.names:
                    target = alias.asname or alias.name
                    imports[target] = alias.name
            elif isinstance(node, ast.ImportFrom):
                import_module = _resolve_import_module(
                    current_module=module,
                    imported_module=node.module,
                    level=int(node.level),
                )
                for alias in node.names:
                    if alias.name == '*':
                        continue
                    target = alias.asname or alias.name
                    if import_module:
                        imports[target] = f"{import_module}.{alias.name}"
                    else:
                        imports[target] = alias.name

        for node in ast.walk(tree):
            if not isinstance(node, ast.ClassDef):
                continue
            full_name = f"{module}.{node.name}" if module else node.name
            bases: list[str] = []
            for base in node.bases:
                raw = _base_name(base)
                if not raw:
                    continue
                bases.append(_resolve_symbol(raw, imports=imports))
            infos.append(
                _ClassInfo(
                    class_name=node.name,
                    full_name=full_name,
                    module=module,
                    path=py_path.as_posix(),
                    lineno=int(node.lineno),
                    resolved_bases=tuple(bases),
                ),
            )
    return infos


def list_struct_classes(*, source_root: Path) -> list[StructClass]:
    infos = _iter_class_infos(source_root)

    by_full_name = {info.full_name: info for info in infos}
    struct_full_names: set[str] = set()

    for info in infos:
        if 'msgspec.Struct' in info.resolved_bases:
            struct_full_names.add(info.full_name)

    changed = True
    while changed:
        changed = False
        for info in infos:
            if info.full_name in struct_full_names:
                continue
            for base in info.resolved_bases:
                if base in struct_full_names:
                    struct_full_names.add(info.full_name)
                    changed = True
                    break
                if '.' not in base:
                    same_module_base = f"{info.module}.{base}" if info.module else base
                    if same_module_base in struct_full_names:
                        struct_full_names.add(info.full_name)
                        changed = True
                        break

    structs: list[StructClass] = []
    for full_name in sorted(struct_full_names):
        info = by_full_name.get(full_name)
        if info is None:
            continue
        structs.append(
            StructClass(
                class_name=info.class_name,
                full_name=info.full_name,
                module=info.module,
                path=info.path,
                lineno=info.lineno,
            ),
        )
    return structs


def _bucket_for_path(path: str) -> str:
    parts = path.split('/')
    if len(parts) >= 3 and parts[0] == 'src' and parts[1] == 'crimson':
        return parts[2]
    if len(parts) >= 2 and parts[0] == 'src':
        return parts[1]
    return 'other'


def summarize_inventory(*, structs: list[StructClass]) -> InventorySummary:
    counts = Counter(_bucket_for_path(item.path) for item in structs)
    by_name: dict[str, list[StructClass]] = defaultdict(list)
    for item in structs:
        by_name[item.class_name].append(item)
    duplicates = {
        name: sorted(items, key=lambda entry: (entry.path, entry.lineno))
        for name, items in by_name.items()
        if len(items) > 1
    }
    return InventorySummary(
        total_structs=len(structs),
        counts_by_bucket=dict(sorted(counts.items(), key=lambda kv: (-kv[1], kv[0]))),
        duplicate_names=dict(sorted(duplicates.items())),
    )


def inventory_as_json(*, summary: InventorySummary, structs: list[StructClass]) -> str:
    payload = {
        'total_structs': int(summary.total_structs),
        'counts_by_bucket': {str(k): int(v) for k, v in summary.counts_by_bucket.items()},
        'duplicate_names': {
            str(name): [
                {
                    'path': item.path,
                    'lineno': int(item.lineno),
                    'module': item.module,
                    'full_name': item.full_name,
                }
                for item in items
            ]
            for name, items in summary.duplicate_names.items()
        },
        'structs': [
            {
                'class_name': item.class_name,
                'full_name': item.full_name,
                'module': item.module,
                'path': item.path,
                'lineno': int(item.lineno),
            }
            for item in structs
        ],
    }
    return json.dumps(payload, indent=2, sort_keys=True)



from __future__ import annotations

import ast
import re
from pathlib import Path

import msgspec

from .channel_helpers import (
    ENTITY_SAMPLE_KINDS,
    checkpoint_channel,
    entity_rows,
    entity_samples_channel,
)
from .payloads import BuiltinObject, BuiltinValue, coerce_builtin_value, to_builtin_object
from .schema import TickRecord
from .trace import TraceReader

_QUERY_RE = re.compile(r"^\s*(ticks|entities)\s+where\s+(.+?)\s*$")
_COND_RE = re.compile(r"^\s*(.+?)\s*(==|!=|>=|<=|>|<)\s*(.+?)\s*$")


class _Operand(msgspec.Struct, frozen=True):
    kind: str
    value: BuiltinValue


class _Predicate(msgspec.Struct, frozen=True):
    left: _Operand
    op: str
    right: _Operand


def _parse_literal(token: str) -> BuiltinValue:
    raw = token.strip()
    lower = raw.lower()
    if lower == "true":
        return True
    if lower == "false":
        return False
    if lower == "none" or lower == "null":
        return None
    try:
        parsed = ast.literal_eval(raw)
    except (SyntaxError, ValueError):
        return raw
    try:
        return coerce_builtin_value(parsed, field="query literal")
    except TypeError:
        return raw


def _parse_operand(token: str) -> _Operand:
    raw = token.strip()
    if raw.startswith("prev(") and raw.endswith(")"):
        path = raw[5:-1].strip()
        if not path:
            raise ValueError("empty prev(...) field path")
        return _Operand(kind="prev_field", value=path)
    parsed = _parse_literal(raw)
    if parsed != raw:
        return _Operand(kind="literal", value=parsed)
    return _Operand(kind="field", value=raw)


def _parse_expression(expression: str) -> tuple[str, _Predicate]:
    scope_match = _QUERY_RE.match(expression)
    if scope_match is None:
        raise ValueError("expression must start with 'ticks where ...' or 'entities where ...'")
    scope = scope_match.group(1)
    condition = scope_match.group(2)
    cond_match = _COND_RE.match(condition)
    if cond_match is None:
        raise ValueError("expression condition must use one of: == != >= <= > <")
    left = _parse_operand(cond_match.group(1))
    op = cond_match.group(2)
    right = _parse_operand(cond_match.group(3))
    return scope, _Predicate(left=left, op=op, right=right)


def _resolve_path(root: BuiltinObject, path: str) -> BuiltinValue | None:
    current: BuiltinValue = root
    for part in path.split("."):
        match current:
            case dict() as mapped:
                current = mapped.get(part)
            case _:
                return None
    return current


def _to_float(value: object) -> float | None:
    match value:
        case bool() as v:
            return float(int(v))
        case int() as v:
            return float(v)
        case float() as v:
            return v
        case str() as v:
            try:
                return float(v)
            except ValueError:
                return None
        case _:
            return None


def _eval_operand(
    operand: _Operand,
    *,
    current: BuiltinObject,
    previous: BuiltinObject | None,
) -> BuiltinValue | None:
    if operand.kind == "literal":
        return operand.value
    if operand.kind == "field":
        return _resolve_path(current, str(operand.value))
    if operand.kind == "prev_field":
        if previous is None:
            return None
        return _resolve_path(previous, str(operand.value))
    raise ValueError(f"unsupported operand kind: {operand.kind}")


def _compare_values(left: BuiltinValue | None, op: str, right: BuiltinValue | None) -> bool:
    left_num = _to_float(left)
    right_num = _to_float(right)
    if left_num is not None and right_num is not None:
        if op == "==":
            return left_num == right_num
        if op == "!=":
            return left_num != right_num
        if op == ">=":
            return left_num >= right_num
        if op == "<=":
            return left_num <= right_num
        if op == ">":
            return left_num > right_num
        if op == "<":
            return left_num < right_num
        raise ValueError(f"unsupported comparison operator: {op}")

    if op == "==":
        return left == right
    if op == "!=":
        return left != right

    left_text = str(left)
    right_text = str(right)
    if op == ">=":
        return left_text >= right_text
    if op == "<=":
        return left_text <= right_text
    if op == ">":
        return left_text > right_text
    if op == "<":
        return left_text < right_text
    raise ValueError(f"unsupported comparison operator: {op}")


def _entity_rows(row: TickRecord) -> list[BuiltinObject]:
    samples = entity_samples_channel(row)
    if samples is None:
        return []
    out: list[BuiltinObject] = []
    for kind in ENTITY_SAMPLE_KINDS:
        for item in entity_rows(samples, kind=kind):
            out.append(to_builtin_object(item, field=f"entity_samples.{kind}"))
    return out


def _event_type_counts(row: TickRecord) -> dict[str, int]:
    checkpoint = checkpoint_channel(row)
    if checkpoint is None:
        return {}
    return {
        "hit_count": int(checkpoint.events.hit_count),
        "pickup_count": int(checkpoint.events.pickup_count),
        "sfx_count": int(checkpoint.events.sfx_count),
    }


def tick_summary_from_row(row: TickRecord) -> BuiltinObject:
    checkpoint_obj = checkpoint_channel(row)
    checkpoint = to_builtin_object(checkpoint_obj, field="checkpoint") if checkpoint_obj is not None else {}
    samples = entity_samples_channel(row)
    if samples is None:
        entity_counts = {kind: 0 for kind in ENTITY_SAMPLE_KINDS}
    else:
        entity_counts = {
            "creatures": len(samples.creatures),
            "projectiles": len(samples.projectiles),
            "secondary_projectiles": len(samples.secondary_projectiles),
            "bonuses": len(samples.bonuses),
        }

    event_counts = _event_type_counts(row)
    event_count_total = sum(event_counts.values())
    event_types_sorted = sorted(event_counts.items(), key=lambda item: (-item[1], item[0]))
    top_event_types = [f"{name}:{count}" for name, count in event_types_sorted[:8]]

    return to_builtin_object(
        {
            "tick_index": row.tick_index,
            "elapsed_ms": row.elapsed_ms,
            "dt_ms_i32": row.dt_ms_i32,
            "mode_id": row.mode_id,
            "phase_markers": list(row.phase_markers),
            "checkpoint": checkpoint,
            "entity_counts": entity_counts,
            "event_count_total": event_count_total,
            "top_event_types": top_event_types,
        },
        field="tick_summary",
    )


def summarize_tick(*, trace_path: Path, tick_index: int) -> BuiltinObject:
    with TraceReader(Path(trace_path)) as trace:
        row = trace.tick(tick_index)
        if row is None:
            raise ValueError(f"tick {tick_index} not found")
        return tick_summary_from_row(row)


def parse_tick_range(value: str | None) -> tuple[int | None, int | None]:
    if value is None:
        return None, None
    raw = value.strip()
    if not raw:
        return None, None
    if ".." not in raw:
        tick = int(raw)
        return tick, tick

    left_raw, right_raw = raw.split("..", maxsplit=1)
    left = int(left_raw) if left_raw.strip() else None
    right = int(right_raw) if right_raw.strip() else None
    if left is not None and right is not None and left > right:
        raise ValueError(f"invalid tick range {value!r}: start > end")
    return left, right


def entity_history(
    *,
    trace_path: Path,
    entity_uid: int,
    tick_start: int | None = None,
    tick_end: int | None = None,
) -> BuiltinObject:
    snapshots: list[BuiltinObject] = []
    spawn_tick: int | None = None
    despawn_tick: int | None = None
    with TraceReader(Path(trace_path)) as trace:
        for row in trace.iter_ticks(tick_start=tick_start, tick_end=tick_end):
            tick_value = int(row.tick_index)
            for entity in _entity_rows(row):
                uid_obj = entity.get("uid")
                if not isinstance(uid_obj, int):
                    continue
                if int(uid_obj) != int(entity_uid):
                    continue
                snapshot = dict(entity)
                snapshot["tick_index"] = tick_value
                snapshots.append(snapshot)
                if spawn_tick is None:
                    spawn_tick = tick_value
                despawn_tick = tick_value

    if not snapshots:
        raise ValueError(f"entity uid {entity_uid} not found in requested range")
    assert spawn_tick is not None
    assert despawn_tick is not None

    first = snapshots[0]
    return to_builtin_object(
        {
            "entity_uid": entity_uid,
            "pool_kind": str(first.get("pool_kind", "unknown")),
            "spawn_tick": int(spawn_tick),
            "despawn_tick": int(despawn_tick),
            "samples": snapshots,
        },
        field="entity_history",
    )


def query_trace(
    *,
    trace_path: Path,
    expression: str,
    limit: int = 256,
) -> BuiltinObject:
    scope, predicate = _parse_expression(expression)
    limit_value = max(1, limit)

    rows: list[BuiltinObject] = []
    matched_count = 0
    with TraceReader(Path(trace_path)) as trace:
        if scope == "ticks":
            prev_context: BuiltinObject | None = None
            for row in trace.iter_ticks():
                context = tick_summary_from_row(row)
                left = _eval_operand(predicate.left, current=context, previous=prev_context)
                right = _eval_operand(predicate.right, current=context, previous=prev_context)
                if _compare_values(left, predicate.op, right):
                    matched_count += 1
                    if len(rows) < limit_value:
                        rows.append(context)
                prev_context = context
        else:
            for row in trace.iter_ticks():
                tick_index = row.tick_index
                for entity in _entity_rows(row):
                    context = {"tick_index": tick_index, **entity}
                    left = _eval_operand(predicate.left, current=context, previous=None)
                    right = _eval_operand(predicate.right, current=context, previous=None)
                    if _compare_values(left, predicate.op, right):
                        matched_count += 1
                        if len(rows) < limit_value:
                            rows.append(context)

    return to_builtin_object(
        {
            "scope": scope,
            "expression": expression,
            "limit": limit_value,
            "match_count": matched_count,
            "truncated": bool(matched_count > len(rows)),
            "rows": rows,
        },
        field="query_trace",
    )



from __future__ import annotations

from enum import Enum
from typing import Final

import msgspec


class SfxId(Enum):
    TROOPER_INPAIN_01 = "sfx_trooper_inpain_01"
    TROOPER_INPAIN_02 = "sfx_trooper_inpain_02"
    TROOPER_INPAIN_03 = "sfx_trooper_inpain_03"
    TROOPER_DIE_01 = "sfx_trooper_die_01"
    TROOPER_DIE_02 = "sfx_trooper_die_02"
    TROOPER_DIE_03 = "sfx_trooper_die_03"
    ZOMBIE_DIE_01 = "sfx_zombie_die_01"
    ZOMBIE_DIE_02 = "sfx_zombie_die_02"
    ZOMBIE_DIE_03 = "sfx_zombie_die_03"
    ZOMBIE_DIE_04 = "sfx_zombie_die_04"
    ZOMBIE_ATTACK_01 = "sfx_zombie_attack_01"
    ZOMBIE_ATTACK_02 = "sfx_zombie_attack_02"
    ALIEN_DIE_01 = "sfx_alien_die_01"
    ALIEN_DIE_02 = "sfx_alien_die_02"
    ALIEN_DIE_03 = "sfx_alien_die_03"
    ALIEN_DIE_04 = "sfx_alien_die_04"
    ALIEN_ATTACK_01 = "sfx_alien_attack_01"
    ALIEN_ATTACK_02 = "sfx_alien_attack_02"
    LIZARD_DIE_01 = "sfx_lizard_die_01"
    LIZARD_DIE_02 = "sfx_lizard_die_02"
    LIZARD_DIE_03 = "sfx_lizard_die_03"
    LIZARD_DIE_04 = "sfx_lizard_die_04"
    LIZARD_ATTACK_01 = "sfx_lizard_attack_01"
    LIZARD_ATTACK_02 = "sfx_lizard_attack_02"
    SPIDER_DIE_01 = "sfx_spider_die_01"
    SPIDER_DIE_02 = "sfx_spider_die_02"
    SPIDER_DIE_03 = "sfx_spider_die_03"
    SPIDER_DIE_04 = "sfx_spider_die_04"
    SPIDER_ATTACK_01 = "sfx_spider_attack_01"
    SPIDER_ATTACK_02 = "sfx_spider_attack_02"
    PISTOL_FIRE = "sfx_pistol_fire"
    PISTOL_RELOAD = "sfx_pistol_reload"
    SHOTGUN_FIRE = "sfx_shotgun_fire"
    SHOTGUN_RELOAD = "sfx_shotgun_reload"
    AUTORIFLE_FIRE = "sfx_autorifle_fire"
    AUTORIFLE_RELOAD = "sfx_autorifle_reload"
    GAUSS_FIRE = "sfx_gauss_fire"
    HRPM_FIRE = "sfx_hrpm_fire"
    SHOCK_FIRE = "sfx_shock_fire"
    PLASMAMINIGUN_FIRE = "sfx_plasmaminigun_fire"
    PLASMASHOTGUN_FIRE = "sfx_plasmashotgun_fire"
    PULSE_FIRE = "sfx_pulse_fire"
    FLAMER_FIRE_01 = "sfx_flamer_fire_01"
    FLAMER_FIRE_02 = "sfx_flamer_fire_02"
    SHOCK_RELOAD = "sfx_shock_reload"
    SHOCK_FIRE_ALT = "sfx_shock_fire_alt"
    SHOCKMINIGUN_FIRE = "sfx_shockminigun_fire"
    ROCKET_FIRE = "sfx_rocket_fire"
    ROCKETMINI_FIRE = "sfx_rocketmini_fire"
    AUTORIFLE_RELOAD_ALT = "sfx_autorifle_reload_alt"
    BULLET_HIT_01 = "sfx_bullet_hit_01"
    BULLET_HIT_02 = "sfx_bullet_hit_02"
    BULLET_HIT_03 = "sfx_bullet_hit_03"
    BULLET_HIT_04 = "sfx_bullet_hit_04"
    BULLET_HIT_05 = "sfx_bullet_hit_05"
    BULLET_HIT_06 = "sfx_bullet_hit_06"
    SHOCK_HIT_01 = "sfx_shock_hit_01"
    EXPLOSION_SMALL = "sfx_explosion_small"
    EXPLOSION_MEDIUM = "sfx_explosion_medium"
    EXPLOSION_LARGE = "sfx_explosion_large"
    SHOCKWAVE = "sfx_shockwave"
    QUESTHIT = "sfx_questhit"
    UI_BONUS = "sfx_ui_bonus"
    UI_BUTTONCLICK = "sfx_ui_buttonclick"
    UI_PANELCLICK = "sfx_ui_panelclick"
    UI_LEVELUP = "sfx_ui_levelup"
    UI_TYPECLICK_01 = "sfx_ui_typeclick_01"
    UI_TYPECLICK_02 = "sfx_ui_typeclick_02"
    UI_TYPEENTER = "sfx_ui_typeenter"
    UI_CLINK_01 = "sfx_ui_clink_01"
    BLOODSPILL_01 = "sfx_bloodspill_01"
    BLOODSPILL_02 = "sfx_bloodspill_02"


class SfxSpec(msgspec.Struct, frozen=True):
    entry_name: str


_SFX_NATIVE_SPECS: Final[tuple[tuple[SfxId, SfxSpec], ...]] = (
    (SfxId.TROOPER_INPAIN_01, SfxSpec("trooper_inPain_01.ogg")),
    (SfxId.TROOPER_INPAIN_02, SfxSpec("trooper_inPain_02.ogg")),
    (SfxId.TROOPER_INPAIN_03, SfxSpec("trooper_inPain_03.ogg")),
    (SfxId.TROOPER_DIE_01, SfxSpec("trooper_die_01.ogg")),
    (SfxId.TROOPER_DIE_02, SfxSpec("trooper_die_02.ogg")),
    (SfxId.TROOPER_DIE_03, SfxSpec("trooper_die_03.ogg")),
    (SfxId.ZOMBIE_DIE_01, SfxSpec("zombie_die_01.ogg")),
    (SfxId.ZOMBIE_DIE_02, SfxSpec("zombie_die_02.ogg")),
    (SfxId.ZOMBIE_DIE_03, SfxSpec("zombie_die_03.ogg")),
    (SfxId.ZOMBIE_DIE_04, SfxSpec("zombie_die_04.ogg")),
    (SfxId.ZOMBIE_ATTACK_01, SfxSpec("zombie_attack_01.ogg")),
    (SfxId.ZOMBIE_ATTACK_02, SfxSpec("zombie_attack_02.ogg")),
    (SfxId.ALIEN_DIE_01, SfxSpec("alien_die_01.ogg")),
    (SfxId.ALIEN_DIE_02, SfxSpec("alien_die_02.ogg")),
    (SfxId.ALIEN_DIE_03, SfxSpec("alien_die_03.ogg")),
    (SfxId.ALIEN_DIE_04, SfxSpec("alien_die_04.ogg")),
    (SfxId.ALIEN_ATTACK_01, SfxSpec("alien_attack_01.ogg")),
    (SfxId.ALIEN_ATTACK_02, SfxSpec("alien_attack_02.ogg")),
    (SfxId.LIZARD_DIE_01, SfxSpec("lizard_die_01.ogg")),
    (SfxId.LIZARD_DIE_02, SfxSpec("lizard_die_02.ogg")),
    (SfxId.LIZARD_DIE_03, SfxSpec("lizard_die_03.ogg")),
    (SfxId.LIZARD_DIE_04, SfxSpec("lizard_die_04.ogg")),
    (SfxId.LIZARD_ATTACK_01, SfxSpec("lizard_attack_01.ogg")),
    (SfxId.LIZARD_ATTACK_02, SfxSpec("lizard_attack_02.ogg")),
    (SfxId.SPIDER_DIE_01, SfxSpec("spider_die_01.ogg")),
    (SfxId.SPIDER_DIE_02, SfxSpec("spider_die_02.ogg")),
    (SfxId.SPIDER_DIE_03, SfxSpec("spider_die_03.ogg")),
    (SfxId.SPIDER_DIE_04, SfxSpec("spider_die_04.ogg")),
    (SfxId.SPIDER_ATTACK_01, SfxSpec("spider_attack_01.ogg")),
    (SfxId.SPIDER_ATTACK_02, SfxSpec("spider_attack_02.ogg")),
    (SfxId.PISTOL_FIRE, SfxSpec("pistol_fire.ogg")),
    (SfxId.PISTOL_RELOAD, SfxSpec("pistol_reload.ogg")),
    (SfxId.SHOTGUN_FIRE, SfxSpec("shotgun_fire.ogg")),
    (SfxId.SHOTGUN_RELOAD, SfxSpec("shotgun_reload.ogg")),
    (SfxId.AUTORIFLE_FIRE, SfxSpec("autorifle_fire.ogg")),
    (SfxId.AUTORIFLE_RELOAD, SfxSpec("autorifle_reload.ogg")),
    (SfxId.GAUSS_FIRE, SfxSpec("gauss_fire.ogg")),
    (SfxId.HRPM_FIRE, SfxSpec("hrpm_fire.ogg")),
    (SfxId.SHOCK_FIRE, SfxSpec("shock_fire.ogg")),
    (SfxId.PLASMAMINIGUN_FIRE, SfxSpec("plasmaMinigun_fire.ogg")),
    (SfxId.PLASMASHOTGUN_FIRE, SfxSpec("plasmaShotgun_fire.ogg")),
    (SfxId.PULSE_FIRE, SfxSpec("pulse_fire.ogg")),
    (SfxId.FLAMER_FIRE_01, SfxSpec("flamer_fire_01.ogg")),
    (SfxId.FLAMER_FIRE_02, SfxSpec("flamer_fire_02.ogg")),
    (SfxId.SHOCK_RELOAD, SfxSpec("shock_reload.ogg")),
    (SfxId.SHOCK_FIRE_ALT, SfxSpec("shock_fire.ogg")),
    (SfxId.SHOCKMINIGUN_FIRE, SfxSpec("shockMinigun_fire.ogg")),
    (SfxId.ROCKET_FIRE, SfxSpec("rocket_fire.ogg")),
    (SfxId.ROCKETMINI_FIRE, SfxSpec("rocketmini_fire.ogg")),
    (SfxId.AUTORIFLE_RELOAD_ALT, SfxSpec("autorifle_reload.ogg")),
    (SfxId.BULLET_HIT_01, SfxSpec("bullet_hit_01.ogg")),
    (SfxId.BULLET_HIT_02, SfxSpec("bullet_hit_02.ogg")),
    (SfxId.BULLET_HIT_03, SfxSpec("bullet_hit_03.ogg")),
    (SfxId.BULLET_HIT_04, SfxSpec("bullet_hit_04.ogg")),
    (SfxId.BULLET_HIT_05, SfxSpec("bullet_hit_05.ogg")),
    (SfxId.BULLET_HIT_06, SfxSpec("bullet_hit_06.ogg")),
    (SfxId.SHOCK_HIT_01, SfxSpec("shock_hit_01.ogg")),
    (SfxId.EXPLOSION_SMALL, SfxSpec("explosion_small.ogg")),
    (SfxId.EXPLOSION_MEDIUM, SfxSpec("explosion_medium.ogg")),
    (SfxId.EXPLOSION_LARGE, SfxSpec("explosion_large.ogg")),
    (SfxId.SHOCKWAVE, SfxSpec("shockwave.ogg")),
    (SfxId.QUESTHIT, SfxSpec("questHit.ogg")),
    (SfxId.UI_BONUS, SfxSpec("ui_bonus.ogg")),
    (SfxId.UI_BUTTONCLICK, SfxSpec("ui_buttonClick.ogg")),
    (SfxId.UI_PANELCLICK, SfxSpec("ui_panelClick.ogg")),
    (SfxId.UI_LEVELUP, SfxSpec("ui_levelUp.ogg")),
    (SfxId.UI_TYPECLICK_01, SfxSpec("ui_typeClick_01.ogg")),
    (SfxId.UI_TYPECLICK_02, SfxSpec("ui_typeClick_02.ogg")),
    (SfxId.UI_TYPEENTER, SfxSpec("ui_typeEnter.ogg")),
    (SfxId.UI_CLINK_01, SfxSpec("ui_clink_01.ogg")),
    (SfxId.BLOODSPILL_01, SfxSpec("bloodSpill_01.ogg")),
    (SfxId.BLOODSPILL_02, SfxSpec("bloodSpill_02.ogg")),
)

SFX_SPECS: Final[dict[SfxId, SfxSpec]] = dict(_SFX_NATIVE_SPECS)

# Extracted from `audio_init_sfx` (FUN_0043caa0) in `crimsonland.exe`.
# `sfx_load_sample()` allocates the first free slot, so the load order defines stable ids.
SFX_NATIVE_ORDER: Final[tuple[SfxId, ...]] = tuple(sfx_id for sfx_id, _spec in _SFX_NATIVE_SPECS)



from __future__ import annotations

import math

from grim.sfx_map import SfxId

from ...projectiles.types import ProjectileTemplateId
from ...rng_caller_static import RngCallerStatic
from ..helpers import perk_active
from ..ids import PerkId
from ..runtime.hook_types import PerkHooks
from ..runtime.player_tick_context import PlayerPerkTickCtx


def tick_man_bomb(ctx: PlayerPerkTickCtx) -> None:
    if not perk_active(ctx.player, PerkId.MAN_BOMB):
        ctx.player.man_bomb_timer = 0.0
        return

    ctx.player.man_bomb_timer += ctx.dt
    if ctx.player.man_bomb_timer > ctx.state.perk_intervals.man_bomb:
        owner = ctx.owner_ref_for_player_projectiles(ctx.state, ctx.player.index)
        for idx in range(8):
            type_id = ProjectileTemplateId.ION_MINIGUN if ((idx & 1) == 0) else ProjectileTemplateId.ION_RIFLE
            caller = (
                RngCallerStatic.PLAYER_UPDATE_MAN_BOMB_ION_MINIGUN_ANGLE
                if type_id is ProjectileTemplateId.ION_MINIGUN
                else RngCallerStatic.PLAYER_UPDATE_MAN_BOMB_ION_RIFLE_ANGLE
            )
            angle = (
                (float(ctx.state.rng.rand(caller=caller) % 50) * 0.01)
                + float(idx) * (math.pi / 4.0)
                - 0.25
            )
            ctx.projectile_spawn(
                ctx.state,
                players=ctx.players,
                pos=ctx.player.pos,
                angle=angle,
                type_id=type_id,
                owner=owner,
                owner_player_index=ctx.player.index,
            )
        ctx.state.sfx_queue.append(SfxId.EXPLOSION_SMALL)

        ctx.player.man_bomb_timer -= ctx.state.perk_intervals.man_bomb
        ctx.state.perk_intervals.man_bomb = 4.0


HOOKS = PerkHooks(
    perk_id=PerkId.MAN_BOMB,
    player_tick_steps=(tick_man_bomb,),
)



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, auto

from grim.config import CrimsonControlsConfig

from ...aim_schemes import AimScheme
from ...movement_controls import MovementControlType


class RebindTarget(Enum):
    PLAYER_MOVE_CODES = auto()
    PLAYER_FIRE_CODE = auto()
    PLAYER_KEYBOARD_AIM_CODES = auto()
    PLAYER_AIM_AXIS_CODES = auto()
    PLAYER_MOVE_AXIS_CODES = auto()
    GLOBAL_PICK_PERK_CODE = auto()
    GLOBAL_RELOAD_CODE = auto()


@dataclass(frozen=True, slots=True)
class RebindRowSpec:
    label: str
    target: RebindTarget
    target_index: int | None = None
    axis: bool = False


def input_configure_for_label(config_id: AimScheme) -> str:
    """Port of `input_configure_for_label` (0x00447c90)."""

    labels = {
        AimScheme.MOUSE: "Mouse",
        AimScheme.KEYBOARD: "Keyboard",
        AimScheme.JOYSTICK: "Joystick",
        AimScheme.MOUSE_RELATIVE: "Mouse relative",
        AimScheme.DUAL_ACTION_PAD: "Dual Action Pad",
        AimScheme.COMPUTER: "Computer",
    }
    return labels.get(config_id, "Unknown")


def input_scheme_label(scheme: MovementControlType) -> str:
    """Port of `input_scheme_label` (0x00447cf0)."""

    labels = {
        MovementControlType.UNKNOWN: "Unknown",
        MovementControlType.RELATIVE: "Relative",
        MovementControlType.STATIC: "Static",
        MovementControlType.DUAL_ACTION_PAD: "Dual Action Pad",
        MovementControlType.MOUSE_POINT_CLICK: "Mouse point click",
        MovementControlType.COMPUTER: "Computer",
    }
    return labels.get(scheme, "Unknown")


def controls_method_labels(controls: CrimsonControlsConfig, *, player_index: int) -> tuple[str, str]:
    player = controls.player(player_index)
    aim_scheme = player.aim_scheme
    move_mode = player.movement
    return input_configure_for_label(aim_scheme), input_scheme_label(move_mode)


def controls_aim_method_dropdown_ids(current_aim_scheme: AimScheme) -> tuple[AimScheme, ...]:
    ids = [
        AimScheme.MOUSE,
        AimScheme.KEYBOARD,
        AimScheme.JOYSTICK,
        AimScheme.MOUSE_RELATIVE,
        AimScheme.DUAL_ACTION_PAD,
    ]
    if current_aim_scheme is AimScheme.COMPUTER:
        # Original menu keeps "Computer" hidden unless loaded from config.
        ids.append(AimScheme.COMPUTER)
    return tuple(ids)


def controls_rebind_plan(
    *,
    aim_scheme: AimScheme,
    move_mode: MovementControlType,
    player_index: int,
) -> tuple[
    tuple[RebindRowSpec, ...],
    tuple[RebindRowSpec, ...],
    tuple[RebindRowSpec, ...],
]:
    """Return (aim_rows, move_rows, misc_rows) for `controls_menu_update`."""

    aim_rows: list[RebindRowSpec] = []
    move_rows: list[RebindRowSpec] = []
    misc_rows: list[RebindRowSpec] = []

    if aim_scheme is AimScheme.KEYBOARD:
        aim_rows.append(RebindRowSpec("Torso left:", RebindTarget.PLAYER_KEYBOARD_AIM_CODES, 0))
        aim_rows.append(RebindRowSpec("Torso right:", RebindTarget.PLAYER_KEYBOARD_AIM_CODES, 1))
    elif aim_scheme is AimScheme.DUAL_ACTION_PAD:
        aim_rows.append(RebindRowSpec("Aim Up/Down Axis:", RebindTarget.PLAYER_AIM_AXIS_CODES, 0, axis=True))
        aim_rows.append(RebindRowSpec("Aim Left/Right Axis:", RebindTarget.PLAYER_AIM_AXIS_CODES, 1, axis=True))
    aim_rows.append(RebindRowSpec("Fire:", RebindTarget.PLAYER_FIRE_CODE))

    if move_mode is MovementControlType.STATIC:
        move_rows.extend(
            (
                RebindRowSpec("Move Up:", RebindTarget.PLAYER_MOVE_CODES, 0),
                RebindRowSpec("Move Down:", RebindTarget.PLAYER_MOVE_CODES, 1),
                RebindRowSpec("Move Left:", RebindTarget.PLAYER_MOVE_CODES, 2),
                RebindRowSpec("Move Right:", RebindTarget.PLAYER_MOVE_CODES, 3),
            ),
        )
    elif move_mode is MovementControlType.RELATIVE:
        move_rows.extend(
            (
                RebindRowSpec("Forward:", RebindTarget.PLAYER_MOVE_CODES, 0),
                RebindRowSpec("Backwards:", RebindTarget.PLAYER_MOVE_CODES, 1),
                RebindRowSpec("Turn left:", RebindTarget.PLAYER_MOVE_CODES, 2),
                RebindRowSpec("Turn right:", RebindTarget.PLAYER_MOVE_CODES, 3),
            ),
        )
    elif move_mode is MovementControlType.DUAL_ACTION_PAD:
        move_rows.extend(
            (
                RebindRowSpec("Up/Down Axis:", RebindTarget.PLAYER_MOVE_AXIS_CODES, 0, axis=True),
                RebindRowSpec("Left/Right Axis:", RebindTarget.PLAYER_MOVE_AXIS_CODES, 1, axis=True),
            ),
        )
    elif move_mode is MovementControlType.MOUSE_POINT_CLICK:
        move_rows.append(RebindRowSpec("Move to cursor:", RebindTarget.GLOBAL_RELOAD_CODE))

    if int(player_index) == 0:
        misc_rows.append(RebindRowSpec("Level Up:", RebindTarget.GLOBAL_PICK_PERK_CODE))
        if move_mode is not MovementControlType.MOUSE_POINT_CLICK:
            misc_rows.append(RebindRowSpec("Reload:", RebindTarget.GLOBAL_RELOAD_CODE))

    return tuple(aim_rows), tuple(move_rows), tuple(misc_rows)



from __future__ import annotations

from typing import TYPE_CHECKING

from ..game_modes import GameMode
from ..perks import PerkId
from ..perks.helpers import perk_active
from ..rng_caller_static import RngCallerStatic
from .ids import BONUS_BY_ID, BonusId

if TYPE_CHECKING:
    from ..gameplay import GameplayState
    from ..sim.state_types import PlayerState
    from .pool import BonusPool


def _bonus_enabled(bonus_id: BonusId) -> bool:
    meta = BONUS_BY_ID.get(bonus_id)
    if meta is None:
        return False
    return meta.bonus_id != BonusId.UNUSED


def _bonus_pick_suppressed(
    *,
    state: GameplayState,
    players: list[PlayerState],
    bonus_id: BonusId,
    has_fire_bullets_drop: bool,
) -> bool:
    if not _bonus_enabled(bonus_id):
        return True
    if state.shock_chain_links_left > 0 and bonus_id == BonusId.SHOCK_CHAIN:
        return True
    if bonus_id == BonusId.FREEZE and state.bonuses.freeze > 0.0:
        return True
    if bonus_id == BonusId.SHIELD and any(player.shield_timer > 0.0 for player in players):
        return True
    if bonus_id == BonusId.WEAPON and has_fire_bullets_drop:
        return True
    if bonus_id == BonusId.WEAPON and any(perk_active(player, PerkId.MY_FAVOURITE_WEAPON) for player in players):
        return True
    if bonus_id == BonusId.MEDIKIT and any(perk_active(player, PerkId.DEATH_CLOCK) for player in players):
        return True
    level = state.quest_level
    if state.game_mode != GameMode.QUESTS or level is None or level.minor != 10:
        return False

    major = level.major
    if bonus_id == BonusId.NUKE:
        return major in (2, 4, 5) or (state.hardcore and major == 3)
    if bonus_id == BonusId.FREEZE:
        return major == 4 or (state.hardcore and major == 2)
    return False


def bonus_pick_random_type(pool: BonusPool, state: GameplayState, players: list[PlayerState]) -> BonusId:
    has_fire_bullets_drop = any(entry.bonus_id == BonusId.FIRE_BULLETS and not entry.picked for entry in pool.entries)

    for _ in range(101):
        roll = state.rng.rand(caller=RngCallerStatic.BONUS_PICK_RANDOM_TYPE_ROLL) % 162 + 1
        # Mirrors `bonus_pick_random_type` (0x412470) mapping:
        # - roll = rand() % 162 + 1  (1..162)
        # - Points: roll 1..13
        # - Energizer: roll 14 with (rand & 0x3F) == 0, else Weapon
        # - Bucketed ids 3..14 via a 10-step loop; if it would exceed 14, returns 0
        #   to force a reroll (matching the `goto LABEL_18` path leaving `v3 == 0`).
        if roll <= 13:
            bonus_id = BonusId.POINTS
        elif roll == 14:
            if (state.rng.rand(caller=RngCallerStatic.BONUS_PICK_RANDOM_TYPE_ENERGIZER) & 0x3F) == 0:
                bonus_id = BonusId.ENERGIZER
            else:
                bonus_id = BonusId.WEAPON
        else:
            bucket_offset = roll - 14
            bonus_value = int(BonusId.WEAPON)
            while bucket_offset > 10:
                bucket_offset -= 10
                bonus_value += 1
                if bonus_value >= 15:
                    bonus_value = int(BonusId.UNUSED)
                    break
            bonus_id = BonusId(bonus_value)
        if bonus_id == BonusId.UNUSED:
            continue
        if _bonus_pick_suppressed(
            state=state,
            players=players,
            bonus_id=bonus_id,
            has_fire_bullets_drop=has_fire_bullets_drop,
        ):
            continue
        return bonus_id
    return BonusId.POINTS



from __future__ import annotations

from pathlib import Path

from grim.assets import RuntimeResources, load_runtime_resources, unload_runtime_resources
from grim.view import View


class RuntimeResourcesView:
    def __init__(self, view: View, *, assets_dir: Path) -> None:
        self._view = view
        self._assets_dir = Path(assets_dir)
        self._resources: RuntimeResources | None = None

    def open(self) -> None:
        self._resources = load_runtime_resources(self._assets_dir)
        try:
            self._view.open()
        except Exception:
            unload_runtime_resources(self._resources)
            self._resources = None
            raise

    def update(self, dt: float) -> None:
        self._view.update(dt)

    def draw(self) -> None:
        self._view.draw()

    def close(self) -> None:
        try:
            self._view.close()
        finally:
            unload_runtime_resources(self._resources)
            self._resources = None



from __future__ import annotations

from typing import cast

import msgspec

from ..replay.types import PackedPlayerInput
from .lockstep_protocol import INPUT_DELAY_TICKS, InputBatch, PauseState, TickFrame
from .lockstep_state import ClientLockstepState, HostLockstepState, HostReadyTick


class ResyncFailureTracker(msgspec.Struct):
    max_failures_per_match: int = 2
    failures: int = 0

    def note_failure(self) -> bool:
        self.failures += 1
        return int(self.failures) >= int(self.max_failures_per_match)

    def reset(self) -> None:
        self.failures = 0


class HostLanAdapter(msgspec.Struct):
    player_count: int
    input_delay_ticks: int = INPUT_DELAY_TICKS
    input_stall_timeout_ms: int = 250
    local_slot_index: int = 0
    lockstep: HostLockstepState = cast(HostLockstepState, None)
    resync_failures: ResyncFailureTracker = msgspec.field(default_factory=ResyncFailureTracker)

    def __post_init__(self) -> None:
        self.lockstep = HostLockstepState(
            player_count=int(self.player_count),
            input_delay_ticks=int(self.input_delay_ticks),
            input_stall_timeout_ms=int(self.input_stall_timeout_ms),
        )

    def submit_local_input(self, *, tick_index: int, packed_input: PackedPlayerInput) -> None:
        self.lockstep.submit_input_sample(
            slot_index=int(self.local_slot_index),
            tick_index=int(tick_index),
            packed_input=list(packed_input),
        )

    def submit_remote_batch(self, batch: InputBatch) -> None:
        self.lockstep.submit_input_batch(batch)

    def emit_ready_frames(
        self,
        *,
        now_ms: int,
    ) -> list[HostReadyTick]:
        return self.lockstep.pop_ready_frames(
            now_ms=int(now_ms),
        )

    def update_pause_state(self, *, now_ms: int) -> PauseState | None:
        return self.lockstep.update_pause_state(now_ms=int(now_ms))


class ClientLanAdapter(msgspec.Struct):
    local_slot_index: int
    input_delay_ticks: int = INPUT_DELAY_TICKS
    input_stall_timeout_ms: int = 250
    lockstep: ClientLockstepState = cast(ClientLockstepState, None)
    resync_failures: ResyncFailureTracker = msgspec.field(default_factory=ResyncFailureTracker)

    def __post_init__(self) -> None:
        self.lockstep = ClientLockstepState(
            local_slot_index=int(self.local_slot_index),
            input_delay_ticks=int(self.input_delay_ticks),
            input_stall_timeout_ms=int(self.input_stall_timeout_ms),
        )

    def queue_local_input(self, packed_input: PackedPlayerInput) -> InputBatch:
        return self.lockstep.queue_local_input(list(packed_input))

    def ingest_tick_frame(self, frame: TickFrame, *, now_ms: int) -> None:
        self.lockstep.ingest_tick_frame(
            frame,
            now_ms=int(now_ms),
        )

    def pop_tick_frame(self) -> TickFrame | None:
        return self.lockstep.pop_canonical_frame()

    def update_pause_state(self, *, now_ms: int) -> PauseState | None:
        return self.lockstep.update_pause_state(now_ms=int(now_ms))



from __future__ import annotations

import html
import importlib
import json
import math
from collections.abc import Sequence
from pathlib import Path
from typing import Any, Protocol, cast

import msgspec


class _TelemetryFrameLike(Protocol):
    tick_index_after_update: int
    frame_ms: float
    update_ms: float
    draw_ms: float
    draw_calls_total: int
    pass_ms: dict[str, float]


class _TelemetryFrame(msgspec.Struct, frozen=True):
    tick_index_after_update: int
    frame_ms: float
    update_ms: float
    draw_ms: float
    draw_calls_total: int
    pass_ms: dict[str, float]


_TIMING_COLORS = ("#4E79A7", "#F28E2B", "#E15759")
_DRAW_CALLS_COLOR = "#59A14F"
_PASS_SCHEME = "tableau20"
_PASS_TOP_N = 8
_PASS_STACK_MAX_POINTS = 1800
_PASS_CHILDREN_BY_PARENT: dict[str, frozenset[str]] = {
    "projectiles_effects": frozenset(
        {
            "laser_sight",
            "primary_projectiles",
            "particle_pool",
            "secondary_projectiles",
            "sprite_effect_pool",
            "effect_pool",
        },
    ),
    "bonus_ui": frozenset(
        {
            "bonus_pickups",
            "bonus_labels",
            "aim_indicators",
            "direction_arrows",
            "aim_enhancements",
        },
    ),
}


class _PassChartMeta(msgspec.Struct, frozen=True):
    pass_names: list[str]
    source_frames: int
    binned_points: int
    bin_size: int


def write_render_telemetry_charts(
    *,
    frames: Sequence[_TelemetryFrameLike],
    out_dir: Path,
    telemetry_json_path: Path | None = None,
) -> dict[str, Path]:
    out = Path(out_dir)
    out.mkdir(parents=True, exist_ok=True)

    frame_timing_path = out / "frame_timing.svg"
    draw_calls_path = out / "draw_calls.svg"
    pass_timing_path = out / "pass_timing_stacked.svg"
    report_path = out / "report.md"

    frame_rows = [_frame_row(frame) for frame in list(frames)]
    _write_frame_timing_svg(frame_timing_path, frame_rows)
    _write_draw_calls_svg(draw_calls_path, frame_rows)
    pass_meta = _write_pass_timing_svg(pass_timing_path, frame_rows)
    _write_report_md(
        report_path,
        telemetry_json_path=telemetry_json_path,
        frame_timing=frame_timing_path,
        draw_calls=draw_calls_path,
        pass_timing=pass_timing_path,
        pass_names=pass_meta.pass_names,
        frame_count=len(frame_rows),
        pass_source_frames=pass_meta.source_frames,
        pass_binned_points=pass_meta.binned_points,
        pass_bin_size=pass_meta.bin_size,
    )

    return {
        "frame_timing_svg": frame_timing_path,
        "draw_calls_svg": draw_calls_path,
        "pass_timing_stacked_svg": pass_timing_path,
        "report_md": report_path,
    }


def write_render_telemetry_charts_from_json(
    *,
    telemetry_json_path: Path,
    out_dir: Path,
) -> dict[str, Path]:
    payload = json.loads(Path(telemetry_json_path).read_text(encoding="utf-8"))
    frames_payload = payload.get("frames")
    if not isinstance(frames_payload, list):
        raise TypeError("telemetry json is missing a frames list")
    frames: list[_TelemetryFrame] = []
    for entry in frames_payload:
        if not isinstance(entry, dict):
            continue
        frame = _TelemetryFrame(
            tick_index_after_update=int(entry.get("tick_index_after_update", 0)),
            frame_ms=float(entry.get("frame_ms", 0.0)),
            update_ms=float(entry.get("update_ms", 0.0)),
            draw_ms=float(entry.get("draw_ms", 0.0)),
            draw_calls_total=int(entry.get("draw_calls_total", 0)),
            pass_ms={str(k): float(v) for k, v in cast(dict[str, Any], entry.get("pass_ms", {})).items()},
        )
        frames.append(frame)
    return write_render_telemetry_charts(
        frames=frames,
        out_dir=out_dir,
        telemetry_json_path=Path(telemetry_json_path),
    )


def _frame_row(frame: _TelemetryFrameLike) -> dict[str, Any]:
    return {
        "tick_index": int(frame.tick_index_after_update),
        "frame_ms": float(frame.frame_ms),
        "update_ms": float(frame.update_ms),
        "draw_ms": float(frame.draw_ms),
        "draw_calls_total": int(frame.draw_calls_total),
        "pass_ms": {str(name): float(value) for name, value in frame.pass_ms.items()},
    }


def _write_frame_timing_svg(path: Path, frames: list[dict[str, Any]]) -> None:
    if not frames:
        _write_empty_svg(path, title="Per-tick frame timing (ms)")
        return

    rows: list[dict[str, Any]] = []
    for frame in frames:
        rows.extend(
            (
                {"tick_index": int(frame["tick_index"]), "metric": "frame_ms", "value": float(frame["frame_ms"])},
                {"tick_index": int(frame["tick_index"]), "metric": "update_ms", "value": float(frame["update_ms"])},
                {"tick_index": int(frame["tick_index"]), "metric": "draw_ms", "value": float(frame["draw_ms"])},
            ),
        )

    alt = _load_altair()
    chart = (
        alt.Chart(alt.Data(values=rows))
        .mark_line(clip=True, strokeWidth=0.5, interpolate="step-after")
        .encode(
            x=alt.X("tick_index:Q", title="Tick"),
            y=alt.Y("value:Q", title="Milliseconds"),
            color=alt.Color(
                "metric:N",
                title="Metric",
                scale=alt.Scale(
                    domain=["frame_ms", "update_ms", "draw_ms"],
                    range=list(_TIMING_COLORS),
                ),
            ),
            tooltip=[
                alt.Tooltip("tick_index:Q", title="Tick"),
                alt.Tooltip("metric:N", title="Metric"),
                alt.Tooltip("value:Q", title="ms", format=".3f"),
            ],
        )
        .properties(width=1180, height=360, title="Per-tick frame timing (ms)")
    )
    _save_svg(path=path, chart=chart)


def _write_draw_calls_svg(path: Path, frames: list[dict[str, Any]]) -> None:
    if not frames:
        _write_empty_svg(path, title="Per-tick draw calls")
        return

    rows = [
        {"tick_index": int(frame["tick_index"]), "draw_calls_total": int(frame["draw_calls_total"])} for frame in frames
    ]
    alt = _load_altair()
    chart = (
        alt.Chart(alt.Data(values=rows))
        .mark_line(clip=True, color=_DRAW_CALLS_COLOR, strokeWidth=0.5, interpolate="step-after")
        .encode(
            x=alt.X("tick_index:Q", title="Tick"),
            y=alt.Y("draw_calls_total:Q", title="Draw calls"),
            tooltip=[
                alt.Tooltip("tick_index:Q", title="Tick"),
                alt.Tooltip("draw_calls_total:Q", title="Draw calls"),
            ],
        )
        .properties(width=1180, height=360, title="Per-tick draw calls")
    )
    _save_svg(path=path, chart=chart)


def _write_pass_timing_svg(path: Path, frames: list[dict[str, Any]]) -> _PassChartMeta:
    if not frames:
        _write_empty_svg(path, title="Per-tick pass timing (stacked exclusive ms)")
        return _PassChartMeta(pass_names=[], source_frames=0, binned_points=0, bin_size=1)

    pass_totals: dict[str, float] = {}
    for frame in frames:
        exclusive_pass_ms = _exclusive_pass_ms(cast(dict[str, float], frame["pass_ms"]))
        for pass_name, pass_ms in exclusive_pass_ms.items():
            pass_totals[str(pass_name)] = pass_totals.get(str(pass_name), 0.0) + float(pass_ms)

    ordered_passes = [name for name, _total in sorted(pass_totals.items(), key=lambda item: item[1], reverse=True)]
    selected_passes = ordered_passes[:_PASS_TOP_N]
    include_other = len(ordered_passes) > len(selected_passes)

    if include_other:
        selected_passes = [*selected_passes, "other"]

    rows, binned_points, bin_size = _build_binned_pass_rows(
        frames=frames,
        selected_passes=selected_passes,
        include_other=include_other,
    )
    if not rows:
        _write_empty_svg(path, title="Per-tick pass timing (stacked exclusive ms)")
        return _PassChartMeta(
            pass_names=selected_passes,
            source_frames=len(frames),
            binned_points=0,
            bin_size=max(1, int(bin_size)),
        )

    alt = _load_altair()
    chart = (
        alt.Chart(alt.Data(values=rows))
        .mark_area(clip=True, opacity=0.9, interpolate="step-after")
        .encode(
            x=alt.X("tick_index:Q", title="Tick"),
            y=alt.Y("pass_ms:Q", title="Milliseconds", stack="zero"),
            color=alt.Color(
                "pass_name:N",
                title="Pass",
                sort=selected_passes,
                scale=alt.Scale(scheme=_PASS_SCHEME),
            ),
            tooltip=[
                alt.Tooltip("tick_index:Q", title="Tick"),
                alt.Tooltip("tick_start:Q", title="Tick start"),
                alt.Tooltip("tick_end:Q", title="Tick end"),
                alt.Tooltip("frames_in_bin:Q", title="Frames"),
                alt.Tooltip("pass_name:N", title="Pass"),
                alt.Tooltip("pass_ms:Q", title="ms", format=".3f"),
            ],
        )
        .properties(width=1180, height=360, title="Per-tick pass timing (stacked exclusive ms)")
    )
    _save_svg(path=path, chart=chart)
    return _PassChartMeta(
        pass_names=selected_passes,
        source_frames=len(frames),
        binned_points=int(binned_points),
        bin_size=int(bin_size),
    )


def _build_binned_pass_rows(
    *,
    frames: list[dict[str, Any]],
    selected_passes: list[str],
    include_other: bool,
) -> tuple[list[dict[str, Any]], int, int]:
    frame_count = len(frames)
    if frame_count <= 0:
        return [], 0, 1

    bin_size = max(1, int(math.ceil(float(frame_count) / float(_PASS_STACK_MAX_POINTS))))
    rows: list[dict[str, Any]] = []
    for start_idx in range(0, frame_count, bin_size):
        chunk = frames[start_idx : start_idx + bin_size]
        if not chunk:
            continue
        frames_in_bin = len(chunk)
        tick_start = int(chunk[0]["tick_index"])
        tick_end = int(chunk[-1]["tick_index"])
        tick_index = (float(tick_start) + float(tick_end)) / 2.0
        sums: dict[str, float] = {name: 0.0 for name in selected_passes}
        for frame in chunk:
            pass_ms = _exclusive_pass_ms(cast(dict[str, float], frame["pass_ms"]))
            other_total = 0.0
            for pass_name, value in pass_ms.items():
                numeric_value = float(value)
                if pass_name in sums:
                    sums[pass_name] += numeric_value
                else:
                    other_total += numeric_value
            if include_other:
                sums["other"] = sums.get("other", 0.0) + other_total
        for pass_name, total in sums.items():
            mean_value = float(total) / float(frames_in_bin)
            if mean_value <= 0.0:
                continue
            rows.append(
                {
                    "tick_index": tick_index,
                    "tick_start": tick_start,
                    "tick_end": tick_end,
                    "frames_in_bin": int(frames_in_bin),
                    "pass_name": pass_name,
                    "pass_ms": mean_value,
                },
            )

    binned_points = int(math.ceil(float(frame_count) / float(bin_size)))
    return rows, binned_points, int(bin_size)


def _exclusive_pass_ms(pass_ms: dict[str, float]) -> dict[str, float]:
    if not pass_ms:
        return {}

    exclusive = {str(name): float(value) for name, value in pass_ms.items()}
    for parent, children in _PASS_CHILDREN_BY_PARENT.items():
        parent_value = exclusive.get(parent)
        if parent_value is None:
            continue
        child_total = 0.0
        for child in children:
            child_total += float(exclusive.get(child, 0.0))
        exclusive[parent] = max(0.0, float(parent_value) - float(child_total))
    return exclusive


def _load_altair() -> Any:
    try:
        alt = importlib.import_module("altair")
    except ModuleNotFoundError as exc:  # pragma: no cover
        raise RuntimeError(
            "render telemetry chart generation requires optional dependency group `charts`; "
            "run `uv sync --extra charts`",
        ) from exc
    alt.data_transformers.disable_max_rows()
    return alt


def _save_svg(*, path: Path, chart: Any) -> None:
    chart.save(str(path))


def _write_empty_svg(path: Path, *, title: str) -> None:
    payload = (
        '<svg xmlns="http://www.w3.org/2000/svg" width="800" height="200" viewBox="0 0 800 200">\n'
        '<rect width="100%" height="100%" fill="#ffffff"/>\n'
        f'<text x="16" y="24" font-size="14" font-family="monospace">{html.escape(title)}</text>\n'
        '<text x="16" y="48" font-size="12" font-family="monospace">No telemetry frames available.</text>\n'
        "</svg>\n"
    )
    path.write_text(payload, encoding="utf-8")


def _write_report_md(
    path: Path,
    *,
    telemetry_json_path: Path | None,
    frame_timing: Path,
    draw_calls: Path,
    pass_timing: Path,
    pass_names: list[str],
    frame_count: int,
    pass_source_frames: int,
    pass_binned_points: int,
    pass_bin_size: int,
) -> None:
    lines = [
        "# Replay Render Telemetry Report",
        "",
        "## Artifacts",
        f"- Frame timing: `{frame_timing.name}`",
        f"- Draw calls: `{draw_calls.name}`",
        f"- Pass timing (stacked, exclusive): `{pass_timing.name}`",
        f"- Frame count: `{int(frame_count)}`",
        f"- Passes shown in stack: `{', '.join(pass_names) if pass_names else 'none'}`",
        (
            f"- Pass stack binning: `{int(pass_source_frames)} source frames -> "
            f"{int(pass_binned_points)} bins (bin_size={int(pass_bin_size)})`"
        ),
    ]
    if telemetry_json_path is not None:
        lines.append(f"- Full telemetry JSON: `{telemetry_json_path}`")
    lines.append("")
    lines.append("Generated with Altair/Vega-Lite. Open the SVG files directly in a browser for inspection.")
    path.write_text("\n".join(lines) + "\n", encoding="utf-8")


__all__ = ["write_render_telemetry_charts", "write_render_telemetry_charts_from_json"]



from __future__ import annotations

import math
from collections.abc import Callable

import msgspec

from grim.color import RGBA
from grim.geom import Vec2
from grim.rand import CrandLike
from grim.sfx_map import SfxId

from ..effects import EffectPool
from ..effects_atlas import EffectId
from ..math_parity import f32
from ..owner_ref import OwnerRef
from ..perks import PerkId
from ..perks.helpers import perk_active
from ..rng_caller_static import RngCallerStatic
from ..sim.state_types import PlayerState
from .damage_types import CreatureDamageType
from .runtime import CREATURE_LIFECYCLE_ALIVE, CreatureState
from .spawn import CreatureFlags, CreatureTypeId


def _any_player_has_perk(players: list[PlayerState], perk_id: PerkId) -> bool:
    return any(perk_active(player, perk_id) for player in players)


class _CreatureDamageCtx(msgspec.Struct):
    creature: CreatureState
    damage: float
    damage_type: int
    impulse: Vec2
    owner: OwnerRef
    dt: float
    players: list[PlayerState]
    rng: CrandLike


_CreatureDamageStep = Callable[[_CreatureDamageCtx], None]


_CREATURE_DEATH_SFX: dict[CreatureTypeId, tuple[SfxId, ...]] = {
    CreatureTypeId.ZOMBIE: (
        SfxId.ZOMBIE_DIE_01,
        SfxId.ZOMBIE_DIE_02,
        SfxId.ZOMBIE_DIE_03,
        SfxId.ZOMBIE_DIE_04,
    ),
    CreatureTypeId.LIZARD: (
        SfxId.LIZARD_DIE_01,
        SfxId.LIZARD_DIE_02,
        SfxId.LIZARD_DIE_03,
        SfxId.LIZARD_DIE_04,
    ),
    CreatureTypeId.ALIEN: (
        SfxId.ALIEN_DIE_01,
        SfxId.ALIEN_DIE_02,
        SfxId.ALIEN_DIE_03,
        SfxId.ALIEN_DIE_04,
    ),
    CreatureTypeId.SPIDER_SP1: (
        SfxId.SPIDER_DIE_01,
        SfxId.SPIDER_DIE_02,
        SfxId.SPIDER_DIE_03,
        SfxId.SPIDER_DIE_04,
    ),
    CreatureTypeId.SPIDER_SP2: (
        SfxId.SPIDER_DIE_01,
        SfxId.SPIDER_DIE_02,
        SfxId.SPIDER_DIE_03,
        SfxId.SPIDER_DIE_04,
    ),
}

_TROOPER_DEATH_SFX: tuple[SfxId, ...] = (
    SfxId.TROOPER_DIE_01,
    SfxId.TROOPER_DIE_02,
    SfxId.TROOPER_DIE_03,
)

# Native `gameplay_reset_state` writes trooper death-bank slots 0..2 only, but
# `creature_apply_damage` still indexes the bank with `rand & 3`. The unwritten
# slot remains BSS-zeroed and resolves to native SFX id 0:
# `sfx_trooper_inpain_01`.
_TROOPER_DEATH_SFX_PRESERVE_BUGS: tuple[SfxId, ...] = (
    *_TROOPER_DEATH_SFX,
    SfxId.TROOPER_INPAIN_01,
)


def _damage_type1_uranium_filled_bullets(ctx: _CreatureDamageCtx) -> None:
    if not _any_player_has_perk(ctx.players, PerkId.URANIUM_FILLED_BULLETS):
        return
    ctx.damage *= 2.0


def _damage_type1_living_fortress(ctx: _CreatureDamageCtx) -> None:
    if not _any_player_has_perk(ctx.players, PerkId.LIVING_FORTRESS):
        return
    for player in ctx.players:
        if float(player.health) <= 0.0:
            continue
        timer = float(player.living_fortress_timer)
        if timer > 0.0:
            ctx.damage *= timer * 0.05 + 1.0


def _damage_type1_barrel_greaser(ctx: _CreatureDamageCtx) -> None:
    if not _any_player_has_perk(ctx.players, PerkId.BARREL_GREASER):
        return
    ctx.damage *= 1.4


def _damage_type1_doctor(ctx: _CreatureDamageCtx) -> None:
    if not _any_player_has_perk(ctx.players, PerkId.DOCTOR):
        return
    ctx.damage *= 1.2


def _damage_type1_heading_jitter(ctx: _CreatureDamageCtx) -> None:
    creature = ctx.creature
    if (creature.flags & CreatureFlags.ANIM_PING_PONG) != 0:
        return
    jitter = float((ctx.rng.rand(caller=RngCallerStatic.CREATURE_APPLY_DAMAGE_HEADING_JITTER) & 0x7F) - 0x40) * 0.002
    size = max(1e-6, float(creature.size))
    turn = jitter / (size * 0.025)
    turn = min(math.pi / 2.0, turn)
    creature.heading += turn


def _damage_type7_ion_gun_master(ctx: _CreatureDamageCtx) -> None:
    if any(perk_active(player, PerkId.ION_GUN_MASTER) for player in ctx.players):
        ctx.damage *= 1.2


def _damage_type4_pyromaniac(ctx: _CreatureDamageCtx) -> None:
    if not _any_player_has_perk(ctx.players, PerkId.PYROMANIAC):
        return
    ctx.damage *= 1.5
    ctx.rng.rand(caller=RngCallerStatic.CREATURE_APPLY_DAMAGE_PYROMANIAC)


def _damage_lethal_ranged_shock_burst(
    *,
    creature: CreatureState,
    rng: CrandLike,
    effects: EffectPool | None,
    detail_preset: int,
) -> None:
    """Port the `creature_apply_damage` lethal branch for `flags & 0x10`."""
    if (creature.flags & CreatureFlags.RANGED_ATTACK_SHOCK) == 0:
        return
    for _ in range(5):
        rotation = float(rng.rand(caller=RngCallerStatic.CREATURE_APPLY_DAMAGE_SHOCK_BURST_ROTATION) & 0x7F) * 0.049087387
        vel = Vec2(
            float((rng.rand(caller=RngCallerStatic.CREATURE_APPLY_DAMAGE_SHOCK_BURST_VEL_X) & 0x7F) - 0x40),
            float((rng.rand(caller=RngCallerStatic.CREATURE_APPLY_DAMAGE_SHOCK_BURST_VEL_Y) & 0x7F) - 0x40),
        )
        scale_step = float(rng.rand(caller=RngCallerStatic.CREATURE_APPLY_DAMAGE_SHOCK_BURST_SCALE_STEP) % 140) * 0.01 + 0.3
        if effects is None:
            continue
        effects.spawn(
            effect_id=int(EffectId.BURST),
            pos=creature.pos,
            vel=vel,
            rotation=rotation,
            scale=1.0,
            half_width=36.0,
            half_height=36.0,
            age=0.0,
            lifetime=0.7,
            flags=0x1D,
            color=RGBA(0.8, 0.8, 0.3, 0.5),
            rotation_step=0.0,
            scale_step=scale_step,
            detail_preset=int(detail_preset),
        )


def resolve_native_death_sfx(
    creature: CreatureState,
    *,
    rng: CrandLike,
    preserve_bugs: bool = False,
) -> tuple[SfxId, ...]:
    """Resolve the native `creature_apply_damage` death sound, if this path owns one."""
    if (creature.flags & CreatureFlags.RANGED_ATTACK_SHOCK) != 0:
        return ()
    roll = rng.rand(caller=RngCallerStatic.CREATURE_APPLY_DAMAGE_DEATH_SFX)
    if creature.type_id == CreatureTypeId.TROOPER:
        if preserve_bugs:
            return (_TROOPER_DEATH_SFX_PRESERVE_BUGS[roll & 3],)
        return (_TROOPER_DEATH_SFX[roll % len(_TROOPER_DEATH_SFX)],)
    options = _CREATURE_DEATH_SFX.get(creature.type_id)
    if options is None:
        return ()
    return (options[roll & 3],)


_CREATURE_DAMAGE_PRE_STEPS: dict[int, tuple[_CreatureDamageStep, ...]] = {
    CreatureDamageType.BULLET: (
        _damage_type1_uranium_filled_bullets,
        _damage_type1_living_fortress,
        _damage_type1_barrel_greaser,
        _damage_type1_doctor,
    ),
}

_CREATURE_DAMAGE_GLOBAL_PRE_STEPS: dict[int, tuple[_CreatureDamageStep, ...]] = {
    CreatureDamageType.ION: (_damage_type7_ion_gun_master,),
}


_CREATURE_DAMAGE_ALIVE_STEPS: dict[int, tuple[_CreatureDamageStep, ...]] = {
    CreatureDamageType.FIRE: (_damage_type4_pyromaniac,),
}


def creature_apply_damage(
    creature: CreatureState,
    *,
    damage_amount: float,
    damage_type: int,
    impulse: Vec2,
    owner: OwnerRef,
    dt: float,
    players: list[PlayerState],
    rng: CrandLike,
    effects: EffectPool | None = None,
    detail_preset: int = 5,
) -> bool:
    """Apply damage to a creature, returning True if the hit killed it.

    This is a partial port of `creature_apply_damage` (FUN_004207c0).

    Notes:
    - Death side-effects are handled by the caller.
    - `damage_type` is a native integer category; call sites must supply it.
    """

    creature.last_hit_owner = owner
    creature.hit_flash_timer = 0.2

    ctx = _CreatureDamageCtx(
        creature=creature,
        damage=float(damage_amount),
        damage_type=int(damage_type),
        impulse=impulse,
        owner=owner,
        dt=float(dt),
        players=players,
        rng=rng,
    )

    for step in _CREATURE_DAMAGE_GLOBAL_PRE_STEPS.get(ctx.damage_type, ()):
        step(ctx)

    for step in _CREATURE_DAMAGE_PRE_STEPS.get(ctx.damage_type, ()):
        step(ctx)
    if ctx.damage_type == CreatureDamageType.BULLET:
        _damage_type1_heading_jitter(ctx)

    if creature.hp <= 0.0:
        if dt > 0.0:
            creature.lifecycle_stage = float(
                f32(float(creature.lifecycle_stage) - float(f32(float(dt) * 15.0))),
            )
        return True

    for step in _CREATURE_DAMAGE_ALIVE_STEPS.get(ctx.damage_type, ()):
        step(ctx)

    creature.hp -= float(ctx.damage)
    creature.vel = creature.vel - ctx.impulse

    if creature.hp <= 0.0:
        if dt > 0.0:
            creature.lifecycle_stage = float(
                f32(float(creature.lifecycle_stage) - float(f32(float(dt)))),
            )
        else:
            creature.lifecycle_stage = float(f32(float(creature.lifecycle_stage) - 0.001))
        creature.vel = creature.vel - impulse * 2.0
        _damage_lethal_ranged_shock_burst(
            creature=creature,
            rng=rng,
            effects=effects,
            detail_preset=int(detail_preset),
        )
        return True

    return False


def creature_apply_damage_with_lethal_followup(
    creature: CreatureState,
    *,
    damage_amount: float,
    damage_type: int,
    impulse: Vec2,
    owner: OwnerRef,
    dt: float,
    players: list[PlayerState],
    rng: CrandLike,
    preserve_bugs: bool = False,
    effects: EffectPool | None = None,
    detail_preset: int = 5,
    on_lethal: Callable[[tuple[SfxId, ...]], None],
) -> bool:
    """Apply damage and run a required lethal follow-up exactly on death transition.

    This helper keeps lethal bookkeeping adjacent to damage application so runtime
    call sites cannot accidentally skip death handling side effects.
    """

    death_start_needed = float(creature.hp) > 0.0 and float(creature.lifecycle_stage) == CREATURE_LIFECYCLE_ALIVE
    killed = creature_apply_damage(
        creature,
        damage_amount=float(damage_amount),
        damage_type=int(damage_type),
        impulse=impulse,
        owner=owner,
        dt=float(dt),
        players=players,
        rng=rng,
        effects=effects,
        detail_preset=int(detail_preset),
    )
    if killed and death_start_needed:
        on_lethal(resolve_native_death_sfx(creature, rng=rng, preserve_bugs=preserve_bugs))
        return True
    return False



from __future__ import annotations

import math

from grim.assets import TextureId
from grim.fonts.small import draw_small_text, measure_small_text_width
from grim.geom import Vec2
from grim.math import clamp
from grim.raylib_api import rl

from ...bonuses import BONUS_BY_ID, BonusId
from ...bonuses.pool import bonus_find_aim_hover_entry, bonus_label_for_entry
from ...weapons import WEAPON_BY_ID, WeaponId
from .constants import _RAD_TO_DEG
from .context import WorldRenderCtx


def bonus_icon_src(texture: rl.Texture, icon_id: int) -> rl.Rectangle:
    grid = 4
    cell_w = float(texture.width) / grid
    cell_h = float(texture.height) / grid
    col = int(icon_id) % grid
    row = int(icon_id) // grid
    return rl.Rectangle(float(col * cell_w), float(row * cell_h), float(cell_w), float(cell_h))


def weapon_icon_src(texture: rl.Texture, icon_index: int) -> rl.Rectangle:
    grid = 8
    cell_w = float(texture.width) / float(grid)
    cell_h = float(texture.height) / float(grid)
    frame = int(icon_index) * 2
    col = frame % grid
    row = frame // grid
    return rl.Rectangle(float(col * cell_w), float(row * cell_h), float(cell_w * 2), float(cell_h))


def bonus_fade(time_left: float, time_max: float) -> float:
    time_left = float(time_left)
    time_max = float(time_max)
    if time_left <= 0.0 or time_max <= 0.0:
        return 0.0
    if time_left < 0.5:
        return clamp(time_left * 2.0, 0.0, 1.0)
    age = time_max - time_left
    if age < 0.5:
        return clamp(age * 2.0, 0.0, 1.0)
    return 1.0


def draw_bonus_pickups(
    render_ctx: WorldRenderCtx,
    *,
    camera: Vec2,
    view_scale: Vec2,
    scale: float,
    alpha: float = 1.0,
) -> None:
    alpha = clamp(float(alpha), 0.0, 1.0)
    if alpha <= 1e-3:
        return
    frame = render_ctx.frame
    resources = frame.resources
    bonuses_texture = resources.texture(TextureId.BONUSES)
    wicons_texture = resources.texture(TextureId.UI_WICONS)

    bubble_src = bonus_icon_src(bonuses_texture, 0)
    bubble_size = 32.0 * scale

    for idx, bonus in enumerate(frame.state.bonus_pool.entries):
        if bonus.bonus_id == BonusId.UNUSED:
            continue

        fade = bonus_fade(float(bonus.time_left), float(bonus.time_max))
        bubble_alpha = clamp(fade * 0.9, 0.0, 1.0) * alpha

        screen = render_ctx._world_to_screen_with(bonus.pos, camera=camera, view_scale=view_scale)
        bubble_dst = rl.Rectangle(screen.x, screen.y, bubble_size, bubble_size)
        bubble_origin = rl.Vector2(bubble_size * 0.5, bubble_size * 0.5)
        tint = rl.Color(255, 255, 255, int(bubble_alpha * 255.0 + 0.5))
        rl.draw_texture_pro(bonuses_texture, bubble_src, bubble_dst, bubble_origin, 0.0, tint)

        bonus_id = bonus.bonus_id
        if bonus_id == BonusId.WEAPON:
            if bonus.amount not in WEAPON_BY_ID:
                continue
            weapon_id = WeaponId(bonus.amount)
            icon_index = int(WEAPON_BY_ID[weapon_id].icon_index)
            if not (0 <= icon_index <= 31):
                continue

            pulse = math.sin(float(frame.bonus_anim_phase)) ** 4 * 0.25 + 0.75
            icon_scale = fade * pulse
            if icon_scale <= 1e-3:
                continue

            src = weapon_icon_src(wicons_texture, icon_index)
            w = 60.0 * icon_scale * scale
            h = 30.0 * icon_scale * scale
            dst = rl.Rectangle(screen.x, screen.y, w, h)
            origin = rl.Vector2(w * 0.5, h * 0.5)
            rl.draw_texture_pro(wicons_texture, src, dst, origin, 0.0, tint)
            continue

        meta = BONUS_BY_ID.get(bonus_id)
        icon_id = int(meta.icon_id) if meta is not None and meta.icon_id is not None else None
        if icon_id is None or icon_id < 0:
            continue
        if bonus_id == BonusId.POINTS and int(bonus.amount) == 1000:
            icon_id += 1

        pulse = math.sin(float(idx) + float(frame.bonus_anim_phase)) ** 4 * 0.25 + 0.75
        icon_scale = fade * pulse
        if icon_scale <= 1e-3:
            continue

        src = bonus_icon_src(bonuses_texture, icon_id)
        size = 32.0 * icon_scale * scale
        rotation_rad = math.sin(float(idx) - float(frame.elapsed_ms) * 0.003) * 0.2
        dst = rl.Rectangle(screen.x, screen.y, size, size)
        origin = rl.Vector2(size * 0.5, size * 0.5)
        rl.draw_texture_pro(bonuses_texture, src, dst, origin, float(rotation_rad * _RAD_TO_DEG), tint)


def draw_bonus_hover_labels(
    render_ctx: WorldRenderCtx,
    *,
    camera: Vec2,
    view_scale: Vec2,
    alpha: float = 1.0,
) -> None:
    alpha = clamp(float(alpha), 0.0, 1.0)
    if alpha <= 1e-3:
        return

    frame = render_ctx.frame
    font = frame.resources.small_font
    text_scale = 1.0
    screen_w = float(rl.get_screen_width())

    shadow = rl.Color(0, 0, 0, int(180 * alpha + 0.5))
    color = rl.Color(230, 230, 230, int(255 * alpha + 0.5))

    for player in frame.players:
        if player.health <= 0.0:
            continue
        hovered = bonus_find_aim_hover_entry(player, frame.state.bonus_pool)
        if hovered is None:
            continue
        _idx, entry = hovered
        label = bonus_label_for_entry(entry, preserve_bugs=bool(frame.state.preserve_bugs))
        if not label:
            continue

        aim = player.aim
        aim_screen = render_ctx._world_to_screen_with(aim, camera=camera, view_scale=view_scale)
        x = aim_screen.x + 16.0
        y = aim_screen.y - 7.0

        if font is not None:
            text_w = measure_small_text_width(font, label)
        else:
            text_w = float(rl.measure_text(label, int(18 * text_scale)))
        if x + text_w > screen_w:
            x = max(0.0, screen_w - text_w)

        if font is not None:
            draw_small_text(font, label, Vec2(x + 1.0, y + 1.0), shadow)
            draw_small_text(font, label, Vec2(x, y), color)
        else:
            rl.draw_text(label, int(x) + 1, int(y) + 1, int(18 * text_scale), shadow)
            rl.draw_text(label, int(x), int(y), int(18 * text_scale), color)



from __future__ import annotations

import msgspec

type BuiltinScalar = None | bool | int | float | str
type BuiltinValue = BuiltinScalar | list[BuiltinValue] | dict[str, BuiltinValue]
type BuiltinObject = dict[str, BuiltinValue]
type BuiltinRows = list[BuiltinObject]


def coerce_builtin_value(value: object, *, field: str) -> BuiltinValue:
    if value is None:
        return None
    if isinstance(value, bool):
        return bool(value)
    if isinstance(value, int):
        return int(value)
    if isinstance(value, float):
        return float(value)
    if isinstance(value, str):
        return str(value)
    if isinstance(value, list):
        return [coerce_builtin_value(item, field=f"{field}[]") for item in value]
    if isinstance(value, tuple):
        return [coerce_builtin_value(item, field=f"{field}[]") for item in value]
    if isinstance(value, dict):
        out: BuiltinObject = {}
        for key, item in value.items():
            if not isinstance(key, str):
                raise TypeError(f"{field} must use string keys, got {type(key).__name__}")
            out[str(key)] = coerce_builtin_value(item, field=f"{field}.{key}")
        return out
    raise TypeError(f"{field} must contain only builtin debug payload values, got {type(value).__name__}")


def to_builtin_value(value: object, *, field: str) -> BuiltinValue:
    return coerce_builtin_value(msgspec.to_builtins(value), field=field)


def to_builtin_object(value: object, *, field: str) -> BuiltinObject:
    built = to_builtin_value(value, field=field)
    if not isinstance(built, dict):
        raise TypeError(f"{field} must convert to a mapping payload")
    return built


def to_builtin_rows(value: object, *, field: str) -> BuiltinRows:
    built = to_builtin_value(value, field=field)
    if not isinstance(built, list):
        raise TypeError(f"{field} must convert to a row list payload")
    rows: BuiltinRows = []
    for index, item in enumerate(built):
        if not isinstance(item, dict):
            raise TypeError(f"{field}[{index}] must be a mapping payload")
        rows.append(item)
    return rows


def builtin_object_or_empty(value: BuiltinValue | None) -> BuiltinObject:
    if not isinstance(value, dict):
        return {}
    try:
        coerced = coerce_builtin_value(value, field="payload")
    except TypeError:
        return {}
    if isinstance(coerced, dict):
        return coerced
    return {}


def builtin_rows_or_empty(value: BuiltinValue | None) -> BuiltinRows:
    if not isinstance(value, list):
        return []
    rows: BuiltinRows = []
    for item in value:
        if not isinstance(item, dict):
            return []
        try:
            rows.append(to_builtin_object(item, field="payload[]"))
        except TypeError:
            return []
    return rows



from __future__ import annotations

from enum import IntEnum

import msgspec

LOCAL_PLAYER_OWNER_ID = -100


class OwnerKind(IntEnum):
    NONE = 0
    PLAYER = 1
    CREATURE = 2


class OwnerRef(msgspec.Struct, frozen=True):
    kind: OwnerKind
    index: int = 0
    local_host: bool = False

    @classmethod
    def none(cls) -> OwnerRef:
        return cls(kind=OwnerKind.NONE)

    @classmethod
    def from_local_player(cls, index: int) -> OwnerRef:
        return cls(kind=OwnerKind.PLAYER, index=int(index), local_host=True)

    @classmethod
    def from_player(cls, index: int) -> OwnerRef:
        return cls(kind=OwnerKind.PLAYER, index=int(index), local_host=False)

    @classmethod
    def from_creature(cls, index: int) -> OwnerRef:
        return cls(kind=OwnerKind.CREATURE, index=int(index), local_host=False)

    @classmethod
    def from_legacy(cls, owner_id: int) -> OwnerRef:
        legacy = int(owner_id)
        if legacy == int(LOCAL_PLAYER_OWNER_ID):
            return cls.from_local_player(0)
        if legacy < 0:
            idx = -1 - legacy
            if idx >= 0:
                return cls.from_player(idx)
            return cls.none()
        return cls.from_creature(legacy)

    def to_legacy(self) -> int:
        if self.kind == OwnerKind.NONE:
            return 0
        if self.kind == OwnerKind.CREATURE:
            return int(self.index)
        if bool(self.local_host) and int(self.index) == 0:
            return int(LOCAL_PLAYER_OWNER_ID)
        return -1 - int(self.index)

    def is_player(self) -> bool:
        return self.kind == OwnerKind.PLAYER

    def player_index(self) -> int | None:
        if self.kind != OwnerKind.PLAYER:
            return None
        return int(self.index)

    def player_index_in_bounds(self, player_count: int) -> int | None:
        idx = self.player_index()
        if idx is None:
            return None
        if 0 <= idx < int(player_count):
            return int(idx)
        return None

    def creature_index(self) -> int | None:
        if self.kind != OwnerKind.CREATURE:
            return None
        return int(self.index)

    def creature_index_in_bounds(self, creature_count: int) -> int | None:
        idx = self.creature_index()
        if idx is None:
            return None
        if 0 <= idx < int(creature_count):
            return int(idx)
        return None

__all__ = [
    "LOCAL_PLAYER_OWNER_ID",
    "OwnerKind",
    "OwnerRef",
]



from __future__ import annotations

import math
import zlib
from typing import Literal

import msgspec

from ..quests.types import SpawnEntry
from ..wire.float32_wire import assert_wire_f32, wire_f32
from .relay_protocol import (
    RESYNC_CHUNK_PAYLOAD_BYTES,
    RESYNC_MAX_SNAPSHOT_BYTES,
    RbResyncBegin,
    RbResyncChunk,
    RbResyncCommit,
)

SCHEMA_VERSION = 4
SNAPSHOT_CODEC = "msgpack_state_v4_f32wire"


class RollbackResyncV5Error(RuntimeError):
    """Raised when rollback resync stream payloads are invalid."""


class ReplayStateSnapshotV2(msgspec.Struct, forbid_unknown_fields=True):
    tick_index: int = 0
    recorded_tick_count: int = 0

    def __post_init__(self) -> None:
        self.tick_index = int(self.tick_index)
        self.recorded_tick_count = int(self.recorded_tick_count)


class SurvivalRuntimeSnapshotV2(msgspec.Struct, forbid_unknown_fields=True):
    elapsed_ms: float = 0.0
    stage: int = 0
    spawn_cooldown_ms: float = 0.0
    perk_pending_count: int = 0

    def __post_init__(self) -> None:
        self.elapsed_ms = wire_f32(float(self.elapsed_ms), field="survival.runtime_state.elapsed_ms")
        self.stage = int(self.stage)
        self.spawn_cooldown_ms = wire_f32(
            float(self.spawn_cooldown_ms),
            field="survival.runtime_state.spawn_cooldown_ms",
        )
        self.perk_pending_count = int(self.perk_pending_count)


class RushRuntimeSnapshotV2(msgspec.Struct, forbid_unknown_fields=True):
    elapsed_ms: float = 0.0
    spawn_cooldown_ms: float = 0.0
    kill_count: int = 0

    def __post_init__(self) -> None:
        self.elapsed_ms = wire_f32(float(self.elapsed_ms), field="rush.runtime_state.elapsed_ms")
        self.spawn_cooldown_ms = wire_f32(float(self.spawn_cooldown_ms), field="rush.runtime_state.spawn_cooldown_ms")
        self.kill_count = int(self.kill_count)


class QuestsRuntimeSnapshotV2(msgspec.Struct, forbid_unknown_fields=True):
    elapsed_ms: float = 0.0
    spawn_entries: tuple[SpawnEntry, ...] = msgspec.field(default_factory=tuple)
    spawn_timeline_ms: float = 0.0
    no_creatures_timer_ms: float = 0.0
    completion_transition_ms: float = 0.0
    perk_pending_count: int = 0

    def __post_init__(self) -> None:
        self.elapsed_ms = wire_f32(float(self.elapsed_ms), field="quests.runtime_state.elapsed_ms")
        self.spawn_entries = tuple(self.spawn_entries)
        self.spawn_timeline_ms = wire_f32(float(self.spawn_timeline_ms), field="quests.runtime_state.spawn_timeline_ms")
        self.no_creatures_timer_ms = wire_f32(
            float(self.no_creatures_timer_ms),
            field="quests.runtime_state.no_creatures_timer_ms",
        )
        self.completion_transition_ms = wire_f32(
            float(self.completion_transition_ms),
            field="quests.runtime_state.completion_transition_ms",
        )
        self.perk_pending_count = int(self.perk_pending_count)


class _ModeStateSnapshotV2Base(msgspec.Struct, forbid_unknown_fields=True):
    schema_version: int = SCHEMA_VERSION
    tick_index: int = 0
    replay_state: ReplayStateSnapshotV2 | None = None

    def __post_init__(self) -> None:
        self.schema_version = int(self.schema_version)
        self.tick_index = int(self.tick_index)


class SurvivalStateSnapshotV2(
    _ModeStateSnapshotV2Base,
    tag="survival",
    tag_field="mode",
):
    runtime_state: SurvivalRuntimeSnapshotV2 = msgspec.field(default_factory=SurvivalRuntimeSnapshotV2)

    @property
    def mode(self) -> Literal["survival"]:
        return "survival"


class RushStateSnapshotV2(
    _ModeStateSnapshotV2Base,
    tag="rush",
    tag_field="mode",
):
    runtime_state: RushRuntimeSnapshotV2 = msgspec.field(default_factory=RushRuntimeSnapshotV2)

    @property
    def mode(self) -> Literal["rush"]:
        return "rush"


class QuestsStateSnapshotV2(
    _ModeStateSnapshotV2Base,
    tag="quests",
    tag_field="mode",
):
    runtime_state: QuestsRuntimeSnapshotV2 = msgspec.field(default_factory=QuestsRuntimeSnapshotV2)

    @property
    def mode(self) -> Literal["quests"]:
        return "quests"


ModeStateSnapshotV2 = SurvivalStateSnapshotV2 | RushStateSnapshotV2 | QuestsStateSnapshotV2


_SNAPSHOT_ENCODER = msgspec.msgpack.Encoder()
_SNAPSHOT_DECODER = msgspec.msgpack.Decoder(type=ModeStateSnapshotV2)


def encode_mode_snapshot(
    *,
    snapshot: ModeStateSnapshotV2,
) -> bytes:
    if int(snapshot.schema_version) != int(SCHEMA_VERSION):
        raise RollbackResyncV5Error("unsupported_snapshot_schema")
    _assert_snapshot_f32(snapshot)
    blob = _SNAPSHOT_ENCODER.encode(snapshot)
    if len(blob) > int(RESYNC_MAX_SNAPSHOT_BYTES):
        raise RollbackResyncV5Error("snapshot_too_large")
    return blob


def decode_mode_snapshot(blob: bytes) -> ModeStateSnapshotV2:
    if len(blob) > int(RESYNC_MAX_SNAPSHOT_BYTES):
        raise RollbackResyncV5Error("snapshot_too_large")
    try:
        snapshot = _SNAPSHOT_DECODER.decode(blob)
    except (msgspec.DecodeError, msgspec.ValidationError) as exc:
        raise RollbackResyncV5Error("snapshot_decode_error") from exc
    if int(snapshot.schema_version) != int(SCHEMA_VERSION):
        raise RollbackResyncV5Error("unsupported_snapshot_schema")
    _assert_snapshot_f32(snapshot)
    return snapshot


def _assert_snapshot_f32(snapshot: ModeStateSnapshotV2) -> None:
    replay_state = snapshot.replay_state
    if replay_state is not None:
        replay_state.tick_index = int(replay_state.tick_index)
        replay_state.recorded_tick_count = int(replay_state.recorded_tick_count)
    match snapshot:
        case SurvivalStateSnapshotV2():
            runtime_state = snapshot.runtime_state
            runtime_state.elapsed_ms = assert_wire_f32(runtime_state.elapsed_ms, field="survival.runtime_state.elapsed_ms")
            runtime_state.spawn_cooldown_ms = assert_wire_f32(
                runtime_state.spawn_cooldown_ms,
                field="survival.runtime_state.spawn_cooldown_ms",
            )
            return
        case RushStateSnapshotV2():
            runtime_state = snapshot.runtime_state
            runtime_state.elapsed_ms = assert_wire_f32(runtime_state.elapsed_ms, field="rush.runtime_state.elapsed_ms")
            runtime_state.spawn_cooldown_ms = assert_wire_f32(
                runtime_state.spawn_cooldown_ms,
                field="rush.runtime_state.spawn_cooldown_ms",
            )
            return
        case QuestsStateSnapshotV2():
            runtime_state = snapshot.runtime_state
            runtime_state.elapsed_ms = assert_wire_f32(runtime_state.elapsed_ms, field="quests.runtime_state.elapsed_ms")
            runtime_state.spawn_entries = tuple(runtime_state.spawn_entries)
            runtime_state.spawn_timeline_ms = assert_wire_f32(
                runtime_state.spawn_timeline_ms,
                field="quests.runtime_state.spawn_timeline_ms",
            )
            runtime_state.no_creatures_timer_ms = assert_wire_f32(
                runtime_state.no_creatures_timer_ms,
                field="quests.runtime_state.no_creatures_timer_ms",
            )
            runtime_state.completion_transition_ms = assert_wire_f32(
                runtime_state.completion_transition_ms,
                field="quests.runtime_state.completion_transition_ms",
            )


class RbResyncMessageSet(msgspec.Struct, frozen=True):
    begin: RbResyncBegin
    chunks: list[RbResyncChunk]
    commit: RbResyncCommit


def build_rb_resync_messages(
    *,
    request_id: str,
    snapshot_tick: int,
    snapshot_blob: bytes,
) -> RbResyncMessageSet:
    payload = bytes(snapshot_blob)
    if len(payload) > int(RESYNC_MAX_SNAPSHOT_BYTES):
        raise RollbackResyncV5Error("snapshot_too_large")

    compressed = zlib.compress(payload, level=6)
    if len(compressed) > int(RESYNC_MAX_SNAPSHOT_BYTES):
        raise RollbackResyncV5Error("compressed_snapshot_too_large")

    chunk_size = max(1, int(RESYNC_CHUNK_PAYLOAD_BYTES))
    total_chunks = max(1, int(math.ceil(len(compressed) / float(chunk_size))))

    begin = RbResyncBegin(
        request_id=str(request_id),
        snapshot_tick=int(snapshot_tick),
        codec=SNAPSHOT_CODEC,
        total_chunks=int(total_chunks),
        compressed_size=len(compressed),
        uncompressed_size=len(payload),
    )

    chunks: list[RbResyncChunk] = []
    for chunk_index in range(int(total_chunks)):
        start = int(chunk_index) * int(chunk_size)
        end = start + int(chunk_size)
        chunks.append(
            RbResyncChunk(
                request_id=str(request_id),
                chunk_index=int(chunk_index),
                payload=bytes(compressed[start:end]),
            ),
        )

    commit = RbResyncCommit(
        request_id=str(request_id),
        snapshot_tick=int(snapshot_tick),
    )
    return RbResyncMessageSet(begin=begin, chunks=chunks, commit=commit)


class RbResyncAssemblerV5(msgspec.Struct):
    max_snapshot_bytes: int = RESYNC_MAX_SNAPSHOT_BYTES
    _begin: RbResyncBegin | None = None
    _chunks: dict[int, bytes] = msgspec.field(default_factory=dict)

    @property
    def request_id(self) -> str:
        begin = self._begin
        if begin is None:
            return ""
        return str(begin.request_id or "")

    def begin(self, message: RbResyncBegin) -> None:
        if str(message.codec) != SNAPSHOT_CODEC:
            raise RollbackResyncV5Error("unsupported_snapshot_codec")
        if int(message.compressed_size) < 0 or int(message.compressed_size) > int(self.max_snapshot_bytes):
            raise RollbackResyncV5Error("compressed_snapshot_size_invalid")
        if int(message.uncompressed_size) < 0 or int(message.uncompressed_size) > int(self.max_snapshot_bytes):
            raise RollbackResyncV5Error("snapshot_size_invalid")
        if int(message.total_chunks) <= 0:
            raise RollbackResyncV5Error("resync_chunk_count_invalid")
        self._begin = message
        self._chunks.clear()

    def push_chunk(self, message: RbResyncChunk) -> None:
        begin = self._begin
        if begin is None:
            raise RollbackResyncV5Error("resync_begin_missing")
        if str(message.request_id or "") != str(begin.request_id or ""):
            raise RollbackResyncV5Error("resync_request_id_mismatch")
        index = int(message.chunk_index)
        if index < 0 or index >= int(begin.total_chunks):
            raise RollbackResyncV5Error("resync_chunk_index_invalid")
        self._chunks[index] = bytes(message.payload)

    def finalize(self, message: RbResyncCommit) -> tuple[int, bytes]:
        begin = self._begin
        if begin is None:
            raise RollbackResyncV5Error("resync_begin_missing")
        if str(message.request_id or "") != str(begin.request_id or ""):
            raise RollbackResyncV5Error("resync_request_id_mismatch")
        if int(message.snapshot_tick) != int(begin.snapshot_tick):
            raise RollbackResyncV5Error("resync_tick_mismatch")

        expected_chunks = int(begin.total_chunks)
        if len(self._chunks) != int(expected_chunks):
            raise RollbackResyncV5Error("resync_chunks_incomplete")
        compressed = b"".join(self._chunks[index] for index in range(int(expected_chunks)))
        if len(compressed) != int(begin.compressed_size):
            raise RollbackResyncV5Error("resync_compressed_size_mismatch")
        if len(compressed) > int(self.max_snapshot_bytes):
            raise RollbackResyncV5Error("compressed_snapshot_too_large")

        payload = zlib.decompress(compressed)
        if len(payload) != int(begin.uncompressed_size):
            raise RollbackResyncV5Error("resync_uncompressed_size_mismatch")
        if len(payload) > int(self.max_snapshot_bytes):
            raise RollbackResyncV5Error("snapshot_too_large")

        tick = int(begin.snapshot_tick)
        self._begin = None
        self._chunks.clear()
        return tick, payload


__all__ = [
    "ModeStateSnapshotV2",
    "QuestsRuntimeSnapshotV2",
    "QuestsStateSnapshotV2",
    "RbResyncAssemblerV5",
    "RbResyncMessageSet",
    "RollbackResyncV5Error",
    "RushRuntimeSnapshotV2",
    "RushStateSnapshotV2",
    "SCHEMA_VERSION",
    "SNAPSHOT_CODEC",
    "SurvivalRuntimeSnapshotV2",
    "SurvivalStateSnapshotV2",
    "ReplayStateSnapshotV2",
    "build_rb_resync_messages",
    "decode_mode_snapshot",
    "encode_mode_snapshot",
]
